from __future__ import annotations

import hashlib
from dataclasses import field
from functools import total_ordering
from typing import Any, Dict, List, Optional, Set, Type, Union

from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import LogicalType

from featuretools.feature_discovery.utils import (
    get_primitive_return_type,
    hash_primitive,
)
from featuretools.primitives.base.primitive_base import PrimitiveBase


@total_ordering
class LiteFeature:
    _name: Optional[str] = None
    _alias: Optional[str] = None

    _logical_type: Optional[Type[LogicalType]] = None
    _tags: Set[str] = field(default_factory=set)
    _primitive: Optional[PrimitiveBase] = None
    _base_features: List[LiteFeature] = field(default_factory=list)
    _df_id: Optional[str] = None

    _id: str
    _n_output_features: int = 1

    _depth = 0
    _related_features: Set[LiteFeature]
    _idx: int = 0

    def __init__(
        self,
        name: Optional[str] = None,
        logical_type: Optional[Type[LogicalType]] = None,
        tags: Optional[Set[str]] = None,
        primitive: Optional[PrimitiveBase] = None,
        base_features: Optional[List[LiteFeature]] = None,
        df_id: Optional[str] = None,
        related_features: Optional[Set[LiteFeature]] = None,
        idx: Optional[int] = None,
    ):
        self._logical_type = logical_type
        self._tags = tags if tags else set()
        self._primitive = primitive
        self._base_features = base_features if base_features else []
        self._df_id = df_id
        self._idx = idx if idx is not None else 0
        self._related_features = related_features if related_features else set()

        if self._primitive:
            if not isinstance(self._primitive, PrimitiveBase):
                raise ValueError("primitive input must be of type PrimitiveBase")

            if len(self.base_features) == 0:
                raise ValueError("there must be base features if given a primitive")

            if self._primitive.commutative:
                self._base_features = sorted(self._base_features)

            self._n_output_features = self._primitive.number_output_features
            self._depth = max([x.depth for x in self.base_features]) + 1

            if name:
                self._alias = name

            self._name = self._primitive.generate_name(
                [x.name for x in self.base_features],
            )

            return_column_schema = get_primitive_return_type(self._primitive)
            self._logical_type = (
                type(return_column_schema.logical_type)
                if return_column_schema.logical_type
                else None
            )

            self._tags = return_column_schema.semantic_tags

        else:
            if name is None:
                raise TypeError("Name must be given if origin feature")

            if self._logical_type is None:
                raise TypeError("Logical Type must be given if origin feature")

            self._name = name

        if self._logical_type is not None and "index" not in self._tags:
            self._tags = self._tags | self._logical_type.standard_tags

        self._id = self._generate_hash()

    @property
    def name(self):
        if self._alias:
            return self._alias
        elif self.is_multioutput():
            return f"{self._name}[{self.idx}]"
        return self._name

    @name.setter
    def name(self, _):
        raise AttributeError("name is immutable")

    def set_alias(self, value: Union[str, None]):
        self._alias = value

    @property
    def non_indexed_name(self):
        if not self.is_multioutput():
            raise ValueError("only used on multioutput features")
        return self._name

    @property
    def logical_type(self):
        return self._logical_type

    @logical_type.setter
    def logical_type(self, _):
        raise AttributeError("logical_type is immutable")

    @property
    def tags(self):
        return self._tags.copy()

    @tags.setter
    def tags(self, _):
        raise AttributeError("tags is immutable")

    @property
    def primitive(self):
        return self._primitive

    @primitive.setter
    def primitive(self, _):
        raise AttributeError("primitive is immutable")

    @property
    def base_features(self):
        return self._base_features

    @base_features.setter
    def base_features(self, _):
        raise AttributeError("base_features are immutable")

    @property
    def df_id(self):
        return self._df_id

    @df_id.setter
    def df_id(self, _):
        raise AttributeError("df_id is immutable")

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, _):
        raise AttributeError("id is immutable")

    @property
    def n_output_features(self):
        return self._n_output_features

    @n_output_features.setter
    def n_output_features(self, _):
        raise AttributeError("n_output_features is immutable")

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, _):
        raise AttributeError("depth is immutable")

    @property
    def related_features(self):
        return self._related_features.copy()

    @related_features.setter
    def related_features(self, value: Set[LiteFeature]):
        self._related_features = value

    @property
    def idx(self):
        return self._idx

    @idx.setter
    def idx(self, _):
        raise AttributeError("idx is immutable")

    @staticmethod
    def hash(
        name: Optional[str],
        primitive: Optional[PrimitiveBase] = None,
        base_features: List[LiteFeature] = [],
        df_id: Optional[str] = None,
        idx: int = 0,
    ):
        hash_msg = hashlib.sha256()

        if primitive:
            # TODO: hashing should be on primitive
            hash_msg.update(hash_primitive(primitive)[0].encode("utf-8"))
            commutative = primitive.commutative
            assert (
                len(base_features) > 0
            ), "there must be base features if give a primitive"
            base_columns = base_features
            if commutative:
                base_features.sort()

            for c in base_columns:
                hash_msg.update(c.id.encode("utf-8"))

        else:
            assert name
            hash_msg.update(name.encode("utf-8"))
            if df_id:
                hash_msg.update(df_id.encode("utf-8"))

        hash_msg.update(str(idx).encode("utf-8"))

        return hash_msg.hexdigest()

    def __eq__(self, other: LiteFeature):
        return self._id == other._id

    def __lt__(self, other: LiteFeature):
        return self._id < other._id

    def __ne__(self, other):
        return self._id != other._id

    def __hash__(self):
        return hash(self._id)

    def _generate_hash(self) -> str:
        return self.hash(
            name=self._name,
            primitive=self._primitive,
            base_features=self._base_features,
            df_id=self._df_id,
            idx=self._idx,
        )

    def get_primitive_name(self) -> Union[str, None]:
        return self._primitive.name if self._primitive else None

    def get_dependencies(self, deep=False) -> List[LiteFeature]:
        flattened_dependencies = []
        for f in self._base_features:
            flattened_dependencies.append(f)

            if deep:
                dependencies = f.get_dependencies()
                if isinstance(dependencies, list):
                    flattened_dependencies.extend(dependencies)
                else:
                    flattened_dependencies.append(dependencies)
        return flattened_dependencies

    def get_origin_features(self) -> List[LiteFeature]:
        all_dependencies = self.get_dependencies(deep=True)
        return [f for f in all_dependencies if f._depth == 0]

    @property
    def column_schema(self) -> ColumnSchema:
        return ColumnSchema(logical_type=self.logical_type, semantic_tags=self.tags)

    def dependent_primitives(self) -> Set[Type[PrimitiveBase]]:
        dependent_features = self.get_dependencies(deep=True)
        dependent_primitives = {
            type(f._primitive) for f in dependent_features if f._primitive
        }
        if self._primitive:
            dependent_primitives.add(type(self._primitive))
        return dependent_primitives

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "logical_type": self.logical_type.__name__ if self.logical_type else None,
            "tags": list(self.tags),
            "primitive": hash_primitive(self.primitive)[0] if self.primitive else None,
            "base_features": [x.id for x in self.base_features],
            "df_id": self.df_id,
            "id": self.id,
            "related_features": [x.id for x in self.related_features],
            "idx": self.idx,
        }

    def is_multioutput(self) -> bool:
        return len(self._related_features) > 0

    def copy(self) -> LiteFeature:
        copied_feature = LiteFeature(
            name=self._name,
            logical_type=self._logical_type,
            tags=self._tags.copy(),
            primitive=self._primitive,
            base_features=[f.copy() for f in self._base_features],
            df_id=self._df_id,
            idx=self._idx,
            related_features=self._related_features.copy(),
        )

        copied_feature.set_alias(self._alias)

        return copied_feature

    def __repr__(self) -> str:
        name = f"name='{self.name}'"
        logical_type = f"logical_type={self.logical_type}"
        tags = f"tags={self.tags}"
        primitive = f"primitive={self.get_primitive_name()}"
        return f"LiteFeature({name}, {logical_type}, {tags}, {primitive})"



from featuretools.utils.gen_utils import import_or_raise


def check_graphviz():
    GRAPHVIZ_ERR_MSG = (
        "Please install graphviz to plot."
        + " (See https://featuretools.alteryx.com/en/stable/install.html#installing-graphviz for"
        + " details)"
    )
    graphviz = import_or_raise("graphviz", GRAPHVIZ_ERR_MSG)
    # Try rendering a dummy graph to see if a working backend is installed
    try:
        graphviz.Digraph().pipe(format="svg")
    except graphviz.backend.ExecutableNotFound:
        raise RuntimeError(
            "To plot entity sets, a graphviz backend is required.\n"
            + "Install the backend using one of the following commands:\n"
            + "  Mac OS: brew install graphviz\n"
            + "  Linux (Ubuntu): $ sudo apt install graphviz\n"
            + "  Windows (conda): conda install -c conda-forge python-graphviz\n"
            + "  Windows (pip): pip install graphviz\n"
            + "  Windows (EXE required if graphviz was installed via pip): https://graphviz.org/download/#windows"
            + "  For more details visit: https://featuretools.alteryx.com/en/stable/install.html#installing-graphviz",
        )
    return graphviz


def get_graphviz_format(graphviz, to_file):
    if to_file:
        # Explicitly cast to str in case a Path object was passed in
        to_file = str(to_file)

        split_path = to_file.split(".")
        if len(split_path) < 2:
            raise ValueError(
                "Please use a file extension like '.pdf'"
                + " so that the format can be inferred",
            )

        format_ = split_path[-1]
        valid_formats = graphviz.FORMATS
        if format_ not in valid_formats:
            raise ValueError(
                "Unknown format. Make sure your format is"
                + " amongst the following: %s" % valid_formats,
            )
    else:
        format_ = None
    return format_


def save_graph(graph, to_file, format_):
    # Graphviz always appends the format to the file name, so we need to
    # remove it manually to avoid file names like 'file_name.pdf.pdf'
    offset = len(format_) + 1  # Add 1 for the dot
    output_path = to_file[:-offset]
    graph.render(output_path, cleanup=True)



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools.primitives.base import AggregationPrimitive


class NUniqueMonths(AggregationPrimitive):
    """Determines the number of unique months.

    Description:
        Given a list of datetimes, return the number of unique months.
        NUniqueMonths counts absolute month, not month of year, so the
        same month in two different years is treated as different. (i.e.
        Feb 2017 is different than Feb 2019.)

    Examples:
        >>> from datetime import datetime
        >>> n_unique_months = NUniqueMonths()
        >>> times = [datetime(2019, 1, 1),
        ...          datetime(2019, 1, 2),
        ...          datetime(2019, 1, 3),
        ...          datetime(2019, 2, 1),
        ...          datetime(2018, 2, 1)]
        >>> n_unique_months(times)
        3
    """

    name = "n_unique_months"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def n_unique_months(x):
            return x.dt.to_period("M").nunique()

        return n_unique_months



# flake8: noqa
from featuretools.version import __version__
from featuretools.config_init import config
from featuretools.entityset.api import *
from featuretools import primitives
from featuretools.synthesis.api import *
from featuretools.primitives import list_primitives, summarize_primitives
from featuretools.computational_backends.api import *
from featuretools import tests
from featuretools.utils.recommend_primitives import get_recommended_primitives
from featuretools.utils.time_utils import *
from featuretools.utils.utils_info import show_info
import featuretools.demo
from featuretools import feature_base
from featuretools import selection
from featuretools.feature_base import (
    AggregationFeature,
    DirectFeature,
    Feature,
    FeatureBase,
    GroupByTransformFeature,
    IdentityFeature,
    TransformFeature,
    graph_feature,
    describe_feature,
    save_features,
    load_features,
)

import logging
import pkg_resources
import sys
import traceback
import warnings
from woodwork import list_logical_types, list_semantic_tags

logger = logging.getLogger("featuretools")

# Call functions registered by other libraries when featuretools is imported
for entry_point in pkg_resources.iter_entry_points("featuretools_initialize"):
    try:
        method = entry_point.load()
        if callable(method):
            method()
    except Exception:
        pass
for entry_point in pkg_resources.iter_entry_points("alteryx_open_src_initialize"):
    try:
        method = entry_point.load()
        if callable(method):
            method("featuretools")
    except Exception:
        pass

# Load in submodules registered by other libraries into Featuretools namespace
for entry_point in pkg_resources.iter_entry_points("featuretools_plugin"):
    try:
        sys.modules["featuretools." + entry_point.name] = entry_point.load()
    except Exception:
        message = "Featuretools failed to load plugin {} from library {}. "
        message += "For a full stack trace, set logging to debug."
        logger.warning(message.format(entry_point.name, entry_point.module_name))
        logger.debug(traceback.format_exc())



import json
import os

import pytest
from woodwork.column_schema import ColumnSchema

from featuretools import describe_feature
from featuretools.feature_base import (
    AggregationFeature,
    DirectFeature,
    GroupByTransformFeature,
    IdentityFeature,
    TransformFeature,
)
from featuretools.primitives import (
    Absolute,
    AggregationPrimitive,
    CumMean,
    EqualScalar,
    Mean,
    Mode,
    NMostCommon,
    NumUnique,
    PercentTrue,
    Sum,
    TransformPrimitive,
)


def test_identity_description(es):
    feature = IdentityFeature(es["log"].ww["session_id"])
    description = 'The "session_id".'

    assert describe_feature(feature) == description


def test_direct_description(es):
    feature = DirectFeature(
        IdentityFeature(es["customers"].ww["loves_ice_cream"]),
        "sessions",
    )
    description = (
        'The "loves_ice_cream" for the instance of "customers" associated '
        'with this instance of "sessions".'
    )
    assert describe_feature(feature) == description

    deep_direct = DirectFeature(feature, "log")
    deep_description = (
        'The "loves_ice_cream" for the instance of "customers" '
        'associated with the instance of "sessions" associated with '
        'this instance of "log".'
    )
    assert describe_feature(deep_direct) == deep_description

    agg = AggregationFeature(
        IdentityFeature(es["log"].ww["purchased"]),
        "sessions",
        PercentTrue,
    )
    complicated_direct = DirectFeature(agg, "log")
    agg_on_direct = AggregationFeature(complicated_direct, "products", Mean)

    complicated_description = (
        "The average of the percentage of true values in "
        'the "purchased" of all instances of "log" for each "id" in "sessions" for '
        'the instance of "sessions" associated with this instance of "log" of all '
        'instances of "log" for each "id" in "products".'
    )
    assert describe_feature(agg_on_direct) == complicated_description


def test_transform_description(es):
    feature = TransformFeature(IdentityFeature(es["log"].ww["value"]), Absolute)
    description = 'The absolute value of the "value".'
    assert describe_feature(feature) == description


def test_groupby_transform_description(es):
    feature = GroupByTransformFeature(
        IdentityFeature(es["log"].ww["value"]),
        CumMean,
        IdentityFeature(es["log"].ww["session_id"]),
    )
    description = 'The cumulative mean of the "value" for each "session_id".'

    assert describe_feature(feature) == description


def test_aggregation_description(es):
    feature = AggregationFeature(
        IdentityFeature(es["log"].ww["value"]),
        "sessions",
        Mean,
    )
    description = 'The average of the "value" of all instances of "log" for each "id" in "sessions".'
    assert describe_feature(feature) == description

    stacked_agg = AggregationFeature(feature, "customers", Sum)
    stacked_description = (
        'The sum of t{} of all instances of "sessions" for each "id" '
        'in "customers".'.format(description[1:-1])
    )
    assert describe_feature(stacked_agg) == stacked_description


def test_aggregation_description_where(es):
    where_feature = TransformFeature(
        IdentityFeature(es["log"].ww["countrycode"]),
        EqualScalar("US"),
    )
    feature = AggregationFeature(
        IdentityFeature(es["log"].ww["value"]),
        "sessions",
        Mean,
        where=where_feature,
    )
    description = (
        'The average of the "value" of all instances of "log" where the '
        '"countrycode" is US for each "id" in "sessions".'
    )

    assert describe_feature(feature) == description


def test_aggregation_description_use_previous(es):
    feature = AggregationFeature(
        IdentityFeature(es["log"].ww["value"]),
        "sessions",
        Mean,
        use_previous="5d",
    )
    description = 'The average of the "value" of the previous 5 days of "log" for each "id" in "sessions".'

    assert describe_feature(feature) == description


def test_multioutput_description(es):
    n_most_common = NMostCommon(2)
    n_most_common_feature = AggregationFeature(
        IdentityFeature(es["log"].ww["zipcode"]),
        "sessions",
        n_most_common,
    )
    first_most_common_slice = n_most_common_feature[0]
    second_most_common_slice = n_most_common_feature[1]

    n_most_common_base = 'The 2 most common values of the "zipcode" of all instances of "log" for each "id" in "sessions".'
    n_most_common_first = (
        'The most common value of the "zipcode" of all instances of "log" '
        'for each "id" in "sessions".'
    )
    n_most_common_second = (
        'The 2nd most common value of the "zipcode" of all instances of '
        '"log" for each "id" in "sessions".'
    )

    assert describe_feature(n_most_common_feature) == n_most_common_base
    assert describe_feature(first_most_common_slice) == n_most_common_first
    assert describe_feature(second_most_common_slice) == n_most_common_second

    class CustomMultiOutput(TransformPrimitive):
        name = "custom_multioutput"
        input_types = [ColumnSchema(semantic_tags={"category"})]
        return_type = ColumnSchema(semantic_tags={"category"})

        number_output_features = 4

    custom_feat = TransformFeature(
        IdentityFeature(es["log"].ww["zipcode"]),
        CustomMultiOutput,
    )

    generic_base = 'The result of applying CUSTOM_MULTIOUTPUT to the "zipcode".'
    generic_first = 'The 1st output from applying CUSTOM_MULTIOUTPUT to the "zipcode".'
    generic_second = 'The 2nd output from applying CUSTOM_MULTIOUTPUT to the "zipcode".'

    assert describe_feature(custom_feat) == generic_base
    assert describe_feature(custom_feat[0]) == generic_first
    assert describe_feature(custom_feat[1]) == generic_second

    CustomMultiOutput.description_template = [
        "the multioutput of {}",
        "the {nth_slice} multioutput part of {}",
    ]
    template_base = 'The multioutput of the "zipcode".'
    template_first_slice = 'The 1st multioutput part of the "zipcode".'
    template_second_slice = 'The 2nd multioutput part of the "zipcode".'
    template_third_slice = 'The 3rd multioutput part of the "zipcode".'
    template_fourth_slice = 'The 4th multioutput part of the "zipcode".'
    assert describe_feature(custom_feat) == template_base
    assert describe_feature(custom_feat[0]) == template_first_slice
    assert describe_feature(custom_feat[1]) == template_second_slice
    assert describe_feature(custom_feat[2]) == template_third_slice
    assert describe_feature(custom_feat[3]) == template_fourth_slice

    CustomMultiOutput.description_template = [
        "the multioutput of {}",
        "the primary multioutput part of {}",
        "the secondary multioutput part of {}",
    ]
    custom_base = 'The multioutput of the "zipcode".'
    custom_first_slice = 'The primary multioutput part of the "zipcode".'
    custom_second_slice = 'The secondary multioutput part of the "zipcode".'
    bad_slice_error = "Slice out of range of template"
    assert describe_feature(custom_feat) == custom_base
    assert describe_feature(custom_feat[0]) == custom_first_slice
    assert describe_feature(custom_feat[1]) == custom_second_slice
    with pytest.raises(IndexError, match=bad_slice_error):
        describe_feature(custom_feat[2])


def test_generic_description(es):
    class NoName(TransformPrimitive):
        input_types = [ColumnSchema(semantic_tags={"category"})]
        output_type = ColumnSchema(semantic_tags={"category"})

        def generate_name(self, base_feature_names):
            return "%s(%s%s)" % (
                "NO_NAME",
                ", ".join(base_feature_names),
                self.get_args_string(),
            )

    class CustomAgg(AggregationPrimitive):
        name = "custom_aggregation"
        input_types = [ColumnSchema(semantic_tags={"category"})]
        output_type = ColumnSchema(semantic_tags={"category"})

    class CustomTrans(TransformPrimitive):
        name = "custom_transform"
        input_types = [ColumnSchema(semantic_tags={"category"})]
        output_type = ColumnSchema(semantic_tags={"category"})

    no_name = TransformFeature(IdentityFeature(es["log"].ww["zipcode"]), NoName)
    no_name_description = 'The result of applying NoName to the "zipcode".'
    assert describe_feature(no_name) == no_name_description

    custom_agg = AggregationFeature(
        IdentityFeature(es["log"].ww["zipcode"]),
        "customers",
        CustomAgg,
    )
    custom_agg_description = 'The result of applying CUSTOM_AGGREGATION to the "zipcode" of all instances of "log" for each "id" in "customers".'
    assert describe_feature(custom_agg) == custom_agg_description

    custom_trans = TransformFeature(
        IdentityFeature(es["log"].ww["zipcode"]),
        CustomTrans,
    )
    custom_trans_description = (
        'The result of applying CUSTOM_TRANSFORM to the "zipcode".'
    )
    assert describe_feature(custom_trans) == custom_trans_description


def test_column_description(es):
    column_description = "the name of the device used for each session"
    es["sessions"].ww.columns["device_name"].description = column_description
    identity_feat = IdentityFeature(es["sessions"].ww["device_name"])
    assert (
        describe_feature(identity_feat)
        == column_description[0].upper() + column_description[1:] + "."
    )


def test_metadata(es, tmp_path):
    identity_feature_descriptions = {
        "sessions: device_name": "the name of the device used for each session",
        "customers: id": "the customer's id",
    }
    agg_feat = AggregationFeature(
        IdentityFeature(es["sessions"].ww["device_name"]),
        "customers",
        NumUnique,
    )
    agg_description = (
        "The number of unique elements in the name of the device used for each "
        'session of all instances of "sessions" for each customer\'s id.'
    )
    assert (
        describe_feature(agg_feat, feature_descriptions=identity_feature_descriptions)
        == agg_description
    )

    transform_feat = GroupByTransformFeature(
        IdentityFeature(es["log"].ww["value"]),
        CumMean,
        IdentityFeature(es["log"].ww["session_id"]),
    )
    transform_description = 'The running average of the "value" for each "session_id".'
    primitive_templates = {"cum_mean": "the running average of {}"}
    assert (
        describe_feature(transform_feat, primitive_templates=primitive_templates)
        == transform_description
    )

    custom_agg = AggregationFeature(
        IdentityFeature(es["log"].ww["zipcode"]),
        "sessions",
        Mode,
    )
    auto_description = 'The most frequently occurring value of the "zipcode" of all instances of "log" for each "id" in "sessions".'
    custom_agg_description = "the most frequently used zipcode"
    custom_feature_description = (
        custom_agg_description[0].upper() + custom_agg_description[1:] + "."
    )
    feature_description_dict = {"sessions: MODE(log.zipcode)": custom_agg_description}
    assert describe_feature(custom_agg) == auto_description
    assert (
        describe_feature(custom_agg, feature_descriptions=feature_description_dict)
        == custom_feature_description
    )

    metadata = {
        "feature_descriptions": {
            **identity_feature_descriptions,
            **feature_description_dict,
        },
        "primitive_templates": primitive_templates,
    }
    metadata_path = os.path.join(tmp_path, "description_metadata.json")
    with open(metadata_path, "w") as f:
        json.dump(metadata, f)
    assert describe_feature(agg_feat, metadata_file=metadata_path) == agg_description
    assert (
        describe_feature(transform_feat, metadata_file=metadata_path)
        == transform_description
    )
    assert (
        describe_feature(custom_agg, metadata_file=metadata_path)
        == custom_feature_description
    )



import pandas as pd
import pytest
from woodwork import list_logical_types as ww_list_logical_types
from woodwork import list_semantic_tags as ww_list_semantic_tags

from featuretools import list_logical_types, list_semantic_tags
from featuretools.utils.gen_utils import (
    camel_and_title_to_snake,
    import_or_none,
    import_or_raise,
    is_instance,
)

dd = import_or_none("dask.dataframe")


def test_import_or_raise_errors():
    with pytest.raises(ImportError, match="error message"):
        import_or_raise("_featuretools", "error message")


def test_import_or_raise_imports():
    math = import_or_raise("math", "error message")
    assert math.ceil(0.1) == 1


def test_import_or_none():
    math = import_or_none("math")
    assert math.ceil(0.1) == 1

    bad_lib = import_or_none("_featuretools")
    assert bad_lib is None


@pytest.fixture
def df():
    return pd.DataFrame({"id": range(5)})


def test_is_instance_single_module(df):
    assert is_instance(df, pd, "DataFrame")


@pytest.mark.skipif("not dd")
def test_is_instance_multiple_modules(df):
    df2 = dd.from_pandas(df, npartitions=2)
    assert is_instance(df, (dd, pd), "DataFrame")
    assert is_instance(df2, (dd, pd), "DataFrame")
    assert is_instance(df2["id"], (dd, pd), ("Series", "DataFrame"))
    assert not is_instance(df2["id"], (dd, pd), ("DataFrame", "Series"))


def test_is_instance_errors_mismatch():
    msg = "Number of modules does not match number of classnames"
    with pytest.raises(ValueError, match=msg):
        is_instance("abc", pd, ("DataFrame", "Series"))


def test_is_instance_none_module(df):
    assert not is_instance(df, None, "DataFrame")
    assert is_instance(df, (None, pd), "DataFrame")
    assert is_instance(df, (None, pd), ("Series", "DataFrame"))


def test_list_logical_types():
    ft_ltypes = list_logical_types()
    ww_ltypes = ww_list_logical_types()
    assert ft_ltypes.equals(ww_ltypes)


def test_list_semantic_tags():
    ft_semantic_tags = list_semantic_tags()
    ww_semantic_tags = ww_list_semantic_tags()
    assert ft_semantic_tags.equals(ww_semantic_tags)


def test_camel_and_title_to_snake():
    assert camel_and_title_to_snake("Top3Words") == "top_3_words"
    assert camel_and_title_to_snake("top3Words") == "top_3_words"
    assert camel_and_title_to_snake("Top100Words") == "top_100_words"
    assert camel_and_title_to_snake("top100Words") == "top_100_words"
    assert camel_and_title_to_snake("Top41") == "top_41"
    assert camel_and_title_to_snake("top41") == "top_41"
    assert camel_and_title_to_snake("41TopWords") == "41_top_words"
    assert camel_and_title_to_snake("TopThreeWords") == "top_three_words"
    assert camel_and_title_to_snake("topThreeWords") == "top_three_words"
    assert camel_and_title_to_snake("top_three_words") == "top_three_words"
    assert camel_and_title_to_snake("over_65") == "over_65"
    assert camel_and_title_to_snake("65_and_over") == "65_and_over"
    assert camel_and_title_to_snake("USDValue") == "usd_value"



import numpy as np
import pandas as pd

from featuretools.primitives import NumberOfHashtags
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestNumberOfHashtags(PrimitiveTestBase):
    primitive = NumberOfHashtags

    def test_regular_input(self):
        x = pd.Series(
            [
                "#hello #hi #hello",
                "#regular#expression#0or1#yes",
                "andorandorand #32309",
            ],
        )
        expected = [3.0, 0.0, 0.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_unicode_input(self):
        x = pd.Series(
            [
                "#Ángel #Æ #ĘÁÊÚ",
                "#############Āndandandandand###",
                "andorandorand #32309",
            ],
        )
        expected = [3.0, 0.0, 0.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_multiline(self):
        x = pd.Series(
            [
                "#\n\t\n",
                "#hashtag\n#hashtag2\n#\n\n",
            ],
        )

        expected = [0.0, 2.0]
        actual = self.primitive().get_function()(x)
        np.testing.assert_array_equal(actual, expected)

    def test_null(self):
        x = pd.Series([np.nan, pd.NA, None, "#test"])

        actual = self.primitive().get_function()(x)
        expected = [np.nan, np.nan, np.nan, 1.0]
        np.testing.assert_array_equal(actual, expected)

    def test_alphanumeric_and_special(self):
        x = pd.Series(["#1or0", "#12", "#??!>@?@#>"])

        actual = self.primitive().get_function()(x)
        expected = [1.0, 0.0, 0.0]
        np.testing.assert_array_equal(actual, expected)

    def test_underscore(self):
        x = pd.Series(["#no", "#__yes", "#??!>@?@#>"])

        actual = self.primitive().get_function()(x)
        expected = [1.0, 1.0, 0.0]
        np.testing.assert_array_equal(actual, expected)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



import numpy as np
import pandas as pd
from pytest import raises

from featuretools.primitives import (
    CountAboveMean,
    CountGreaterThan,
    CountInsideNthSTD,
    CountInsideRange,
    CountLessThan,
    CountOutsideNthSTD,
    CountOutsideRange,
)
from featuretools.tests.primitive_tests.utils import PrimitiveTestBase


class TestCountAboveMean(PrimitiveTestBase):
    primitive = CountAboveMean

    def test_regular(self):
        data = pd.Series([1, 2, 3, 4, 5])
        expected = 2
        primitive_func = self.primitive().get_function()
        actual = primitive_func(data)
        assert expected == actual

        data = pd.Series([1, 2, 3.1, 4, 5])
        expected = 3
        primitive_func = self.primitive().get_function()
        actual = primitive_func(data)
        assert expected == actual

    def test_nan_without_ignore_nan(self):
        data = pd.Series([np.nan, 1, 2, 3, 4, 5, np.nan, np.nan])
        expected = np.nan

        primitive_func = self.primitive(skipna=False).get_function()
        actual = primitive_func(data)
        assert np.isnan(actual) == np.isnan(expected)

        data = pd.Series([np.nan])
        primitive_func = self.primitive(skipna=False).get_function()
        actual = primitive_func(data)
        assert np.isnan(actual) == np.isnan(expected)

    def test_nan_with_ignore_nan(self):
        data = pd.Series([np.nan, 1, 2, 3, 4, 5, np.nan, np.nan])
        expected = 2
        primitive_func = self.primitive(skipna=True).get_function()
        actual = primitive_func(data)
        assert expected == actual

        data = pd.Series([np.nan, 1, 2, 3.1, 4, 5, np.nan, np.nan])
        expected = 3
        primitive_func = self.primitive(skipna=True).get_function()
        actual = primitive_func(data)
        assert expected == actual

        data = pd.Series([np.nan])
        expected = np.nan
        primitive_func = self.primitive(skipna=True).get_function()
        actual = primitive_func(data)
        assert np.isnan(actual) == np.isnan(expected)

    def test_inf(self):
        data = pd.Series([np.NINF, 1, 2, 3, 4, 5])
        expected = 5
        primitive_func = self.primitive().get_function()
        actual = primitive_func(data)
        assert expected == actual

        data = pd.Series([1, 2, 3, 4, 5, np.inf])
        expected = 0
        primitive_func = self.primitive().get_function()
        actual = primitive_func(data)
        assert expected == actual

        data = pd.Series([np.NINF, 1, 2, 3, 4, 5, np.inf])
        expected = np.nan
        primitive_func = self.primitive().get_function()
        actual = primitive_func(data)
        assert np.isnan(actual) == np.isnan(expected)

        primitive_func = self.primitive(skipna=False).get_function()
        actual = primitive_func(data)
        assert np.isnan(actual) == np.isnan(expected)


class TestCountGreaterThan(PrimitiveTestBase):
    primitive = CountGreaterThan

    def compare_results(self, data, thresholds, results):
        for threshold, result in zip(thresholds, results):
            primitive = self.primitive(threshold=threshold)
            function = primitive.get_function()
            assert function(data) == result
            assert isinstance(function(data), np.int64)

    def test_regular(self):
        data = pd.Series([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
        thresholds = pd.Series([-5, -2, 0, 2, 5])
        results = pd.Series([10, 7, 5, 3, 0])
        self.compare_results(data, thresholds, results)

    def test_edges(self):
        data = pd.Series([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
        thresholds = pd.Series([np.inf, np.NINF, None, np.nan])
        results = pd.Series([0, len(data), 0, 0])
        self.compare_results(data, thresholds, results)

    def test_nans(self):
        data = pd.Series([-5, -4, -3, np.inf, np.NINF, np.nan, 1, 2, 3, 4, 5])
        thresholds = pd.Series([np.inf, np.NINF, None, 0, np.nan])
        results = pd.Series([0, 9, 0, 6, 0])
        self.compare_results(data, thresholds, results)


class TestCountInsideNthSTD:
    primitive = CountInsideNthSTD

    def test_normal_distribution(self):
        x = pd.Series(
            [
                -76.0,
                41.0,
                -43.0,
                -152.0,
                -89.0,
                28.0,
                49.0,
                298.0,
                -132.0,
                146.0,
                -107.0,
                -26.0,
                26.0,
                -81.0,
                116.0,
                -217.0,
                -102.0,
                144.0,
                120.0,
                -130.0,
            ],
        )

        first_outliers = [-152.0, 298.0, 146.0, 116.0, -217.0, 144.0, 120.0]
        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - len(first_outliers)

        second_outliers = [298.0]
        primitive_instance = self.primitive(2)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - len(second_outliers)

    def test_poisson_distribution(self):
        x = pd.Series(
            [
                1,
                1,
                3,
                3,
                0,
                0,
                1,
                3,
                3,
                1,
                2,
                3,
                2,
                0,
                1,
                3,
                2,
                1,
                0,
                2,
            ],
        )

        first_outliers = [3, 3, 0, 0, 3, 3, 3, 0, 3, 0]
        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - len(first_outliers)

        second_outliers = []
        primitive_instance = self.primitive(2)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - len(second_outliers)

    def test_nan(self):
        # test if function ignores nan values
        x = pd.Series(
            [
                -76.0,
                41.0,
                -43.0,
                -152.0,
                -89.0,
                28.0,
                49.0,
                298.0,
                -132.0,
                146.0,
                -107.0,
                -26.0,
                26.0,
                -81.0,
                116.0,
                -217.0,
                -102.0,
                144.0,
                120.0,
                -130.0,
            ],
        )
        x = pd.concat([x, pd.Series([np.nan] * 20)])
        first_outliers = [-152.0, 298.0, 146.0, 116.0, -217.0, 144.0, 120.0]
        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - len(first_outliers) - 20

        # test a series with all nan values
        x = pd.Series([np.nan] * 20)

        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 0

    def test_negative_n(self):
        with raises(ValueError):
            self.primitive(-1)


class TestCountInsideRange(PrimitiveTestBase):
    primitive = CountInsideRange

    def test_integer_range(self):
        # all integers from -100 to 100
        x = pd.Series(np.arange(-100, 101, 1))
        primitive_instance = self.primitive(-100, 100)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 201

        primitive_instance = self.primitive(-50, 50)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 101

        primitive_instance = self.primitive(1, 1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 1

    def test_float_range(self):
        x = pd.Series(np.linspace(-3, 3, 10))

        primitive_instance = self.primitive(-3, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 10

        primitive_instance = self.primitive(-0.34, 1.68)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 4

        primitive_instance = self.primitive(-3, -3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 1

    def test_nan(self):
        x = pd.Series(np.linspace(-3, 3, 10))
        x = pd.concat([x, pd.Series([np.nan] * 20)])

        primitive_instance = self.primitive(-0.34, 1.68)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 4

        primitive_instance = self.primitive(-3, 3, False)
        primitive_func = primitive_instance.get_function()
        assert np.isnan(primitive_func(x))

    def test_inf(self):
        x = pd.Series(np.linspace(-3, 3, 10))
        num_NINF = 20
        x = pd.concat([x, pd.Series([np.NINF] * num_NINF)])
        num_inf = 10
        x = pd.concat([x, pd.Series([np.inf] * num_inf)])

        primitive_instance = self.primitive(-3, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 10

        primitive_instance = self.primitive(np.NINF, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 10 + num_NINF

        primitive_instance = self.primitive(-3, np.inf)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 10 + num_inf


class TestCountLessThan(PrimitiveTestBase):
    primitive = CountLessThan

    def compare_answers(self, data, thresholds, answers):
        for threshold, answer in zip(thresholds, answers):
            primitive = self.primitive(threshold=threshold)
            function = primitive.get_function()
            assert function(data) == answer
            assert isinstance(function(data), np.int64)

    def test_regular(self):
        data = pd.Series([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
        thresholds = pd.Series([-5, -2, 0, 2, 5])
        answers = pd.Series([0, 3, 5, 7, 10])
        self.compare_answers(data, thresholds, answers)

    def test_edges(self):
        data = pd.Series([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
        thresholds = pd.Series([np.inf, np.NINF, None, np.nan])
        answers = pd.Series([len(data), 0, 0, 0])
        self.compare_answers(data, thresholds, answers)

    def test_nans(self):
        data = pd.Series([-5, -4, -3, np.inf, np.NINF, np.nan, 1, 2, 3, 4, 5])
        thresholds = pd.Series([np.inf, np.NINF, None, 0, np.nan])
        answers = pd.Series([9, 0, 0, 4, 0])
        self.compare_answers(data, thresholds, answers)


class TestCountOutsideNthSTD(PrimitiveTestBase):
    primitive = CountOutsideNthSTD

    def test_normal_distribution(self):
        x = pd.Series(
            [
                10,
                386,
                479,
                627,
                20,
                523,
                482,
                483,
                542,
                699,
                535,
                617,
                577,
                471,
                615,
                583,
                441,
                562,
                563,
                527,
                453,
                530,
                433,
                541,
                585,
                704,
                443,
                569,
                430,
                637,
                331,
                511,
                552,
                496,
                484,
                566,
                554,
                472,
                335,
                440,
                579,
                341,
                545,
                615,
                548,
                604,
                439,
                556,
                442,
                461,
                624,
                611,
                444,
                578,
                405,
                487,
                490,
                496,
                398,
                512,
                422,
                455,
                449,
                432,
                607,
                679,
                434,
                597,
                639,
                565,
                415,
                486,
                668,
                414,
                665,
                763,
                557,
                304,
                404,
                454,
                689,
                610,
                483,
                441,
                657,
                590,
                492,
                476,
                437,
                483,
                529,
                363,
                711,
                543,
            ],
        )
        outliers = [10, 20, 763]
        primitive_instance = self.primitive(2)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(outliers)

    def test_poisson_distribution(self):
        x = pd.Series(
            [
                1,
                1,
                3,
                3,
                0,
                0,
                1,
                3,
                3,
                1,
                2,
                3,
                2,
                0,
                1,
                3,
                2,
                1,
                0,
                2,
            ],
        )

        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 10

        primitive_instance = self.primitive(2)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 0

    def test_nan(self):
        # test if function ignores nan values
        x = pd.Series(
            [
                -76.0,
                41.0,
                -43.0,
                -152.0,
                -89.0,
                28.0,
                49.0,
                298.0,
                -132.0,
                146.0,
                -107.0,
                -26.0,
                26.0,
                -81.0,
                116.0,
                -217.0,
                -102.0,
                144.0,
                120.0,
                -130.0,
            ],
        )
        x = pd.concat([x, pd.Series([np.nan * 20])])
        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 7

        # test a series with all nan values
        x = pd.Series([np.nan] * 20)

        primitive_instance = self.primitive(1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 0

    def test_negative_n(self):
        with raises(ValueError):
            self.primitive(-1)


class TestCountOutsideRange(PrimitiveTestBase):
    primitive = CountOutsideRange

    def test_integer_range(self):
        # all integers from -100 to 100
        x = pd.Series(np.arange(-100, 101, 1))
        primitive_instance = CountOutsideRange(-100, 100)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 0

        primitive_instance = CountOutsideRange(-50, 50)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 100

        primitive_instance = CountOutsideRange(1, 1)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == len(x) - 1

    def test_float_range(self):
        x = pd.Series(np.linspace(-3, 3, 10))

        primitive_instance = CountOutsideRange(-3, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 0

        primitive_instance = CountOutsideRange(-0.34, 1.68)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 6

        primitive_instance = CountOutsideRange(-3, -3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 9

    def test_nan(self):
        x = pd.Series(np.linspace(-3, 3, 10))
        x = pd.concat([x, pd.Series([np.nan] * 20)])
        primitive_instance = CountOutsideRange(-0.34, 1.68)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 6

        primitive_instance = CountOutsideRange(-3, 3, False)
        primitive_func = primitive_instance.get_function()
        assert np.isnan(primitive_func(x))

    def test_inf(self):
        x = pd.Series(np.linspace(-3, 3, 10))
        num_NINF = 20
        x = pd.concat([x, pd.Series([np.NINF] * num_NINF)])
        num_inf = 10
        x = pd.concat([x, pd.Series([np.inf] * num_inf)])

        primitive_instance = CountOutsideRange(-3, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == num_inf + num_NINF

        primitive_instance = CountOutsideRange(-0.34, 1.68)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == 6 + num_inf + num_NINF

        primitive_instance = CountOutsideRange(np.NINF, 3)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == num_inf

        primitive_instance = CountOutsideRange(-3, np.inf)
        primitive_func = primitive_instance.get_function()
        assert primitive_func(x) == num_NINF



from typing import Optional, Tuple

import holidays
import pandas as pd


class HolidayUtil:
    def __init__(self, country="US"):
        try:
            country, subdivision = self.convert_to_subdivision(country)
            self.holidays = holidays.country_holidays(
                country=country,
                subdiv=subdivision,
            )
        except NotImplementedError:
            available_countries = (
                "https://github.com/dr-prodigy/python-holidays#available-countries"
            )
            error = "must be one of the available countries:\n%s" % available_countries
            raise ValueError(error)

        self.federal_holidays = getattr(holidays, country)(years=range(1950, 2075))

    def to_df(self):
        holidays_df = pd.DataFrame(
            sorted(self.federal_holidays.items()),
            columns=["holiday_date", "names"],
        )
        holidays_df.holiday_date = holidays_df.holiday_date.astype("datetime64[ns]")
        return holidays_df

    def convert_to_subdivision(self, country: str) -> Tuple[str, Optional[str]]:
        """Convert country to country + subdivision

           Created in response to library changes that changed countries to subdivisions

        Args:
            country (str): Original country name

        Returns:
            Tuple[str,Optional[str]]: country, subdivsion
        """
        return {
            "ENGLAND": ("GB", country),
            "NORTHERNIRELAND": ("GB", country),
            "PORTUGALEXT": ("PT", "Ext"),
            "PTE": ("PT", "Ext"),
            "SCOTLAND": ("GB", country),
            "UK": ("GB", country),
            "WALES": ("GB", country),
        }.get(country.upper(), (country, None))



import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import CityblockDistance, GeoMidpoint, IsInGeoBox


def test_cityblock():
    primitive_instance = CityblockDistance()
    latlong_1 = pd.Series([(i, i) for i in range(3)])
    latlong_2 = pd.Series([(i, i) for i in range(3, 6)])
    answer = pd.Series([414.56051391, 414.52893691, 414.43421555])
    given_answer = primitive_instance(latlong_1, latlong_2)
    np.testing.assert_allclose(given_answer, answer, rtol=1e-09)

    primitive_instance = CityblockDistance(unit="kilometers")
    answer = primitive_instance(latlong_1, latlong_2)
    given_answer = pd.Series([667.1704814, 667.11966315, 666.96722389])
    np.testing.assert_allclose(given_answer, answer, rtol=1e-09)


def test_cityblock_nans():
    primitive_instance = CityblockDistance()
    lats_longs_1 = [(i, i) for i in range(2)]
    lats_longs_2 = [(i, i) for i in range(2, 4)]
    lats_longs_1 += [(1, 1), (np.nan, 3), (4, np.nan), (np.nan, np.nan)]
    lats_longs_2 += [(np.nan, np.nan), (np.nan, 5), (6, np.nan), (np.nan, np.nan)]
    given_answer = pd.Series(list([276.37367594, 276.35262728] + [np.nan] * 4))
    answer = primitive_instance(lats_longs_1, lats_longs_2)
    np.testing.assert_allclose(given_answer, answer, rtol=1e-09)


def test_cityblock_error():
    error_text = "Invalid unit given"
    with pytest.raises(ValueError, match=error_text):
        CityblockDistance(unit="invalid")


def test_midpoint():
    latlong1 = pd.Series([(-90, -180), (90, 180)])
    latlong2 = pd.Series([(+90, +180), (-90, -180)])
    function = GeoMidpoint().get_function()
    answer = function(latlong1, latlong2)
    for lat, longi in answer:
        assert lat == 0.0
        assert longi == 0.0


def test_midpoint_floating():
    latlong1 = pd.Series([(-45.5, -100.5), (45.5, 100.5)])
    latlong2 = pd.Series([(+45.5, +100.5), (-45.5, -100.5)])
    function = GeoMidpoint().get_function()
    answer = function(latlong1, latlong2)
    for lat, longi in answer:
        assert lat == 0.0
        assert longi == 0.0


def test_midpoint_zeros():
    latlong1 = pd.Series([(0, 0), (0, 0)])
    latlong2 = pd.Series([(0, 0), (0, 0)])
    function = GeoMidpoint().get_function()
    answer = function(latlong1, latlong2)
    for lat, longi in answer:
        assert lat == 0.0
        assert longi == 0.0


def test_midpoint_nan():
    all_nan = pd.Series([(np.nan, np.nan), (np.nan, np.nan)])
    latlong1 = pd.Series([(0, 0), (0, 0)])
    function = GeoMidpoint().get_function()
    answer = function(all_nan, latlong1)
    for lat, longi in answer:
        assert np.isnan(lat)
        assert np.isnan(longi)


def test_isingeobox():
    latlong = pd.Series(
        [
            (1, 2),
            (5, 7),
            (-5, 4),
            (2, 3),
            (0, 0),
            (np.nan, np.nan),
            (-2, np.nan),
            (np.nan, 1),
        ],
    )
    bottomleft = (-5, -5)
    topright = (5, 5)
    primitive = IsInGeoBox(bottomleft, topright)
    function = primitive.get_function()
    primitive_answer = function(latlong)
    answer = pd.Series([True, False, True, True, True, False, False, False])
    assert np.array_equal(primitive_answer, answer)


def test_boston():
    NYC = (40.7128, -74.0060)
    SF = (37.7749, -122.4194)
    Somerville = (42.3876, -71.0995)
    Bejing = (39.9042, 116.4074)
    CapeTown = (-33.9249, 18.4241)
    latlong = pd.Series([NYC, SF, Somerville, Bejing, CapeTown])
    LynnMA = (42.4668, -70.9495)
    DedhamMA = (42.2436, -71.1677)
    primitive = IsInGeoBox(LynnMA, DedhamMA)
    function = primitive.get_function()
    primitive_answer = function(latlong)
    answer = pd.Series([False, False, True, False, False])
    assert np.array_equal(primitive_answer, answer)



import importlib.util
import os
from inspect import getfullargspec, getsource, isclass
from typing import Dict, List

import pandas as pd
from woodwork import list_logical_types, list_semantic_tags, type_system
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import NaturalLanguage

import featuretools
from featuretools.primitives import NumberOfCommonWords
from featuretools.primitives.base import (
    AggregationPrimitive,
    PrimitiveBase,
    TransformPrimitive,
)
from featuretools.utils.gen_utils import Library, find_descendents


def _get_primitives(primitive_kind):
    """Helper function that selects all primitives
    that are instances of `primitive_kind`
    """
    primitives = set()
    for attribute_string in dir(featuretools.primitives):
        attribute = getattr(featuretools.primitives, attribute_string)
        if isclass(attribute):
            if issubclass(attribute, primitive_kind) and attribute.name:
                primitives.add(attribute)
    return {prim.name.lower(): prim for prim in primitives}


def get_aggregation_primitives():
    """Returns all aggregation primitives, regardless
    of compatibility
    """
    return _get_primitives(featuretools.primitives.AggregationPrimitive)


def get_transform_primitives():
    """Returns all transform primitives, regardless
    of compatibility
    """
    return _get_primitives(featuretools.primitives.TransformPrimitive)


def get_all_primitives():
    """Helper function to return all primitives"""
    primitives = set()
    for attribute_string in dir(featuretools.primitives):
        attribute = getattr(featuretools.primitives, attribute_string)
        if isclass(attribute):
            if issubclass(attribute, PrimitiveBase) and attribute.name:
                primitives.add(attribute)
    return {prim.__name__: prim for prim in primitives}


def _get_natural_language_primitives():
    """Returns all Natural Language transform primitives,
    regardless of compatibility
    """
    transform_primitives = get_transform_primitives()

    def _natural_language_in_input_type(primitive):
        for input_type in primitive.input_types:
            if isinstance(input_type, list):
                if any(
                    isinstance(column_schema.logical_type, NaturalLanguage)
                    for column_schema in input_type
                ):
                    return True
            else:
                if isinstance(input_type.logical_type, NaturalLanguage):
                    return True
        return False

    return {
        name: primitive
        for name, primitive in transform_primitives.items()
        if _natural_language_in_input_type(primitive)
    }


def list_primitives():
    """Returns a DataFrame that lists and describes each built-in primitive."""
    trans_names, trans_primitives, valid_inputs, return_type = _get_names_primitives(
        get_transform_primitives,
    )
    trans_dask = [
        Library.DASK in primitive.compatibility for primitive in trans_primitives
    ]
    trans_spark = [
        Library.SPARK in primitive.compatibility for primitive in trans_primitives
    ]
    transform_df = pd.DataFrame(
        {
            "name": trans_names,
            "description": _get_descriptions(trans_primitives),
            "dask_compatible": trans_dask,
            "spark_compatible": trans_spark,
            "valid_inputs": valid_inputs,
            "return_type": return_type,
        },
    )
    transform_df["type"] = "transform"

    agg_names, agg_primitives, valid_inputs, return_type = _get_names_primitives(
        get_aggregation_primitives,
    )
    agg_dask = [Library.DASK in primitive.compatibility for primitive in agg_primitives]
    agg_spark = [
        Library.SPARK in primitive.compatibility for primitive in agg_primitives
    ]
    agg_df = pd.DataFrame(
        {
            "name": agg_names,
            "description": _get_descriptions(agg_primitives),
            "dask_compatible": agg_dask,
            "spark_compatible": agg_spark,
            "valid_inputs": valid_inputs,
            "return_type": return_type,
        },
    )
    agg_df["type"] = "aggregation"

    columns = [
        "name",
        "type",
        "dask_compatible",
        "spark_compatible",
        "description",
        "valid_inputs",
        "return_type",
    ]
    return pd.concat([agg_df, transform_df], ignore_index=True)[columns]


def summarize_primitives() -> pd.DataFrame:
    """Returns a metrics summary DataFrame of all primitives found in list_primitives."""
    (
        trans_names,
        trans_primitives,
        trans_valid_inputs,
        trans_return_type,
    ) = _get_names_primitives(get_transform_primitives)

    (
        agg_names,
        agg_primitives,
        agg_valid_inputs,
        agg_return_type,
    ) = _get_names_primitives(get_aggregation_primitives)

    tot_trans = len(trans_names)
    tot_agg = len(agg_names)
    tot_prims = tot_trans + tot_agg
    all_primitives = trans_primitives + agg_primitives
    primitives_summary = _get_summary_primitives(all_primitives)
    summary_dict = {
        "total_primitives": tot_prims,
        "aggregation_primitives": tot_agg,
        "transform_primitives": tot_trans,
        **primitives_summary["general_metrics"],
    }
    summary_dict.update(
        {
            f"uses_{ltype}_input": count
            for ltype, count in primitives_summary["logical_type_input_metrics"].items()
        },
    )
    summary_dict.update(
        {
            f"uses_{tag}_tag_input": count
            for tag, count in primitives_summary["semantic_tag_metrics"].items()
        },
    )
    summary_df = pd.DataFrame(
        [{"Metric": k, "Count": v} for k, v in summary_dict.items()],
    )
    return summary_df


def get_default_aggregation_primitives():
    agg_primitives = [
        featuretools.primitives.Sum,
        featuretools.primitives.Std,
        featuretools.primitives.Max,
        featuretools.primitives.Skew,
        featuretools.primitives.Min,
        featuretools.primitives.Mean,
        featuretools.primitives.Count,
        featuretools.primitives.PercentTrue,
        featuretools.primitives.NumUnique,
        featuretools.primitives.Mode,
    ]
    return agg_primitives


def get_default_transform_primitives():
    # featuretools.primitives.TimeSince
    trans_primitives = [
        featuretools.primitives.Age,
        featuretools.primitives.Day,
        featuretools.primitives.Year,
        featuretools.primitives.Month,
        featuretools.primitives.Weekday,
        featuretools.primitives.Haversine,
        featuretools.primitives.NumWords,
        featuretools.primitives.NumCharacters,
    ]
    return trans_primitives


def _get_descriptions(primitives):
    descriptions = []
    for prim in primitives:
        description = ""
        if prim.__doc__ is not None:
            # Break on the empty line between the docstring description and the remainder of the docstring
            description = prim.__doc__.split("\n\n")[0]
            # remove any excess whitespace from line breaks
            description = " ".join(description.split())
        descriptions.append(description)
    return descriptions


def _get_summary_primitives(primitives: List) -> Dict[str, int]:
    """Provides metrics for a list of primitives."""
    unique_input_types = set()
    unique_output_types = set()
    uses_multi_input = 0
    uses_multi_output = 0
    uses_external_data = 0
    are_controllable = 0
    logical_type_metrics = {
        log_type: 0 for log_type in list(list_logical_types()["type_string"])
    }
    semantic_tag_metrics = {
        sem_tag: 0 for sem_tag in list(list_semantic_tags()["name"])
    }
    semantic_tag_metrics.update(
        {"foreign_key": 0},
    )  # not currently in list_semantic_tags()

    for prim in primitives:
        log_in_type_checks = set()
        sem_tag_type_checks = set()
        input_types = prim.flatten_nested_input_types(prim.input_types)
        _check_input_types(
            input_types,
            log_in_type_checks,
            sem_tag_type_checks,
            unique_input_types,
        )
        for ltype in list(log_in_type_checks):
            logical_type_metrics[ltype] += 1

        for sem_tag in list(sem_tag_type_checks):
            semantic_tag_metrics[sem_tag] += 1

        if len(prim.input_types) > 1:
            uses_multi_input += 1

        # checks if number_output_features is set as an instance variable or set as a constant
        if (
            "self.number_output_features =" in getsource(prim.__init__)
            or prim.number_output_features > 1
        ):
            uses_multi_output += 1
        unique_output_types.add(str(prim.return_type))

        if hasattr(prim, "filename"):
            uses_external_data += 1

        if len(getfullargspec(prim.__init__).args) > 1:
            are_controllable += 1

    return {
        "general_metrics": {
            "unique_input_types": len(unique_input_types),
            "unique_output_types": len(unique_output_types),
            "uses_multi_input": uses_multi_input,
            "uses_multi_output": uses_multi_output,
            "uses_external_data": uses_external_data,
            "are_controllable": are_controllable,
        },
        "logical_type_input_metrics": logical_type_metrics,
        "semantic_tag_metrics": semantic_tag_metrics,
    }


def _check_input_types(
    input_types: List[ColumnSchema],
    log_in_type_checks: set,
    sem_tag_type_checks: set,
    unique_input_types: set,
):
    """Checks if any logical types or semantic tags occur in a list of Woodwork input types and keeps track of unique input types."""
    for in_type in input_types:
        if in_type.semantic_tags:
            for sem_tag in in_type.semantic_tags:
                sem_tag_type_checks.add(sem_tag)
        if in_type.logical_type:
            log_in_type_checks.add(in_type.logical_type.type_string)
        unique_input_types.add(str(in_type))


def _get_names_primitives(primitive_func):
    names = []
    primitives = []
    valid_inputs = []
    return_type = []
    for name, primitive in primitive_func().items():
        names.append(name)
        primitives.append(primitive)
        input_types = _get_unique_input_types(primitive.input_types)
        valid_inputs.append(", ".join(input_types))
        return_type.append(
            str(primitive.return_type),
        ) if primitive.return_type is not None else return_type.append(None)
    return names, primitives, valid_inputs, return_type


def _get_unique_input_types(input_types):
    types = set()
    for input_type in input_types:
        if isinstance(input_type, list):
            types |= _get_unique_input_types(input_type)
        else:
            types.add(str(input_type))
    return types


def list_primitive_files(directory):
    """returns list of files in directory that might contain primitives"""
    files = os.listdir(directory)
    keep = []
    for path in files:
        if not check_valid_primitive_path(path):
            continue
        keep.append(os.path.join(directory, path))
    return keep


def check_valid_primitive_path(path):
    if os.path.isdir(path):
        return False

    filename = os.path.basename(path)

    if filename[:2] == "__" or filename[0] == "." or filename[-3:] != ".py":
        return False

    return True


def load_primitive_from_file(filepath):
    """load primitive objects in a file"""
    module = os.path.basename(filepath)[:-3]
    # TODO: what is the first argument"?
    spec = importlib.util.spec_from_file_location(module, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    primitives = []
    for primitive_name in vars(module):
        primitive_class = getattr(module, primitive_name)
        if (
            isclass(primitive_class)
            and issubclass(primitive_class, PrimitiveBase)
            and primitive_class not in (AggregationPrimitive, TransformPrimitive)
        ):
            primitives.append((primitive_name, primitive_class))

    if len(primitives) == 0:
        raise RuntimeError("No primitive defined in file %s" % filepath)
    elif len(primitives) > 1:
        raise RuntimeError("More than one primitive defined in file %s" % filepath)

    return primitives[0]


def serialize_primitive(primitive: PrimitiveBase):
    """build a dictionary with the data necessary to construct the given primitive"""
    args_dict = {name: val for name, val in primitive.get_arguments()}
    cls = type(primitive)
    if cls == NumberOfCommonWords and "word_set" in args_dict:
        args_dict["word_set"] = list(args_dict["word_set"])
    return {
        "type": cls.__name__,
        "module": cls.__module__,
        "arguments": args_dict,
    }


class PrimitivesDeserializer(object):
    """
    This class wraps a cache and a generator which iterates over all primitive
    classes. When deserializing a primitive if it is not in the cache then we
    iterate until it is found, adding every seen class to the cache. When
    deserializing the next primitive the iteration resumes where it left off. This
    means that we never visit a class more than once.
    """

    def __init__(self):
        # Cache to avoid repeatedly searching for primitive class
        # (class_name, module_name) -> class
        self.class_cache = {}

        self.primitive_classes = find_descendents(PrimitiveBase)

    def deserialize_primitive(self, primitive_dict):
        """
        Construct a primitive from the given dictionary (output from
        serialize_primitive).
        """
        class_name = primitive_dict["type"]
        module_name = primitive_dict["module"]
        class_cache_key = (class_name, module_name.split(".")[0])

        if class_cache_key in self.class_cache:
            cls = self.class_cache[class_cache_key]
        else:
            cls = self._find_class_in_descendants(class_cache_key)

        if not cls:
            raise RuntimeError(
                'Primitive "%s" in module "%s" not found' % (class_name, module_name),
            )
        arguments = primitive_dict["arguments"]
        if cls == NumberOfCommonWords and "word_set" in arguments:
            # We converted word_set from a set to a list to make it serializable,
            # we should convert it back now.
            arguments["word_set"] = set(arguments["word_set"])
        primitive_instance = cls(**arguments)

        return primitive_instance

    def _find_class_in_descendants(self, search_key):
        for cls in self.primitive_classes:
            cls_key = (cls.__name__, cls.__module__.split(".")[0])
            self.class_cache[cls_key] = cls

            if cls_key == search_key:
                return cls


def get_all_logical_type_names():
    """Helper function that returns all registered woodwork logical types"""
    return {lt.__name__: lt for lt in type_system.registered_types}



from datetime import datetime, timedelta

import numpy as np
import pandas as pd


def make_temporal_cutoffs(
    instance_ids,
    cutoffs,
    window_size=None,
    num_windows=None,
    start=None,
):
    """Makes a set of equally spaced cutoff times prior to a set of input cutoffs and instance ids.

    If window_size and num_windows are provided, then num_windows of size window_size will be created
    prior to each cutoff time

    If window_size and a start list is provided, then a variable number of windows will be created prior
    to each cutoff time, with the corresponding start time as the first cutoff.

    If num_windows and a start list is provided, then num_windows of variable size will be created prior
    to each cutoff time, with the corresponding start time as the first cutoff

    Args:
        instance_ids (list, np.ndarray, or pd.Series): list of instance ids. This function will make a
            new datetime series of multiple cutoff times for each value in this array.
        cutoffs (list, np.ndarray, or pd.Series): list of datetime objects associated with each instance id.
            Each one of these will be the last time in the new datetime series for each instance id
        window_size (pd.Timedelta, optional): amount of time between each datetime in each new cutoff series
        num_windows (int, optional): number of windows in each new cutoff series
        start (list, optional): list of start times for each instance id
    """
    if window_size is not None and num_windows is not None and start is not None:
        raise ValueError(
            "Only supply 2 of the 3 optional args, window_size, num_windows and start",
        )
    out = []
    for i, id_time in enumerate(zip(instance_ids, cutoffs)):
        _id, time = id_time
        _window_size = window_size
        _start = None
        if start is not None:
            if window_size is None:
                _window_size = (time - start[i]) / (num_windows - 1)
            else:
                _start = start[i]
        to_add = pd.DataFrame()
        to_add["time"] = pd.date_range(
            end=time,
            periods=num_windows,
            freq=_window_size,
            start=_start,
        )
        to_add["instance_id"] = [_id] * len(to_add["time"])
        out.append(to_add)
    return pd.concat(out).reset_index(drop=True)


def convert_time_units(secs, unit):
    """
    Converts a time specified in seconds to a time in the given units

    Args:
        secs (integer): number of seconds. This function will convert the units of this number.
        unit(str): units to be converted to.
            acceptable values: years, months, days, hours, minutes, seconds, milliseconds, nanoseconds
    """
    unit_divs = {
        "years": 31540000,
        "months": 2628000,
        "days": 86400,
        "hours": 3600,
        "minutes": 60,
        "seconds": 1,
        "milliseconds": 0.001,
        "nanoseconds": 0.000000001,
    }
    if unit not in unit_divs:
        raise ValueError("Invalid unit given, make sure it is plural")

    return secs / (unit_divs[unit])


def convert_datetime_to_floats(x):
    first = int(x.iloc[0].value * 1e-9)
    x = pd.to_numeric(x).astype(np.float64).values
    dividend = find_dividend_by_unit(first)
    x *= 1e-9 / dividend
    return x


def convert_timedelta_to_floats(x):
    first = int(x.iloc[0].total_seconds())
    dividend = find_dividend_by_unit(first)
    x = pd.TimedeltaIndex(x).total_seconds().astype(np.float64) / dividend
    return x


def find_dividend_by_unit(time):
    """Finds whether time best corresponds to a value in
    days, hours, minutes, or seconds.
    """
    for dividend in [86400, 3600, 60]:
        div = time / dividend
        if round(div) == div:
            return dividend
    return 1


def calculate_trend(series):
    # numpy can't handle `Int64` values, so cast to float
    if series.dtype == "Int64":
        series = series.astype("float64")
    df = pd.DataFrame({"x": series.index, "y": series.values}).dropna()
    if df.shape[0] <= 2:
        return np.nan
    if isinstance(df["x"].iloc[0], (datetime, pd.Timestamp)):
        x = convert_datetime_to_floats(df["x"])
    else:
        x = df["x"].values

    if isinstance(df["y"].iloc[0], (datetime, pd.Timestamp)):
        y = convert_datetime_to_floats(df["y"])
    elif isinstance(df["y"].iloc[0], (timedelta, pd.Timedelta)):
        y = convert_timedelta_to_floats(df["y"])
    else:
        y = df["y"].values

    x = x - x.mean()
    y = y - y.mean()

    # prevent divide by zero error
    if len(np.unique(x)) == 1:
        return 0

    # consider scipy.stats.linregress for large n cases
    coefficients = np.polyfit(x, y, 1)
    return coefficients[0]



from unittest.mock import patch

import composeml as cp
import numpy as np
import pandas as pd
import pytest
from packaging.version import parse
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import NaturalLanguage

from featuretools.computational_backends.calculate_feature_matrix import (
    FEATURE_CALCULATION_PERCENTAGE,
)
from featuretools.entityset import EntitySet, Timedelta
from featuretools.exceptions import UnusedPrimitiveWarning
from featuretools.primitives import GreaterThanScalar, Max, Mean, Min, Sum
from featuretools.primitives.base import AggregationPrimitive, TransformPrimitive
from featuretools.synthesis import dfs
from featuretools.synthesis.deep_feature_synthesis import DeepFeatureSynthesis
from featuretools.tests.testing_utils import to_pandas
from featuretools.utils.gen_utils import Library, import_or_none, is_instance

dd = import_or_none("dask.dataframe")


@pytest.fixture
def datetime_es():
    cards_df = pd.DataFrame({"id": [1, 2, 3, 4, 5]})
    transactions_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5],
            "card_id": [1, 1, 5, 1, 5],
            "transaction_time": pd.to_datetime(
                [
                    "2011-2-28 04:00",
                    "2012-2-28 05:00",
                    "2012-2-29 06:00",
                    "2012-3-1 08:00",
                    "2014-4-1 10:00",
                ],
            ),
            "fraud": [True, False, False, False, True],
        },
    )

    datetime_es = EntitySet(id="fraud_data")
    datetime_es = datetime_es.add_dataframe(
        dataframe_name="transactions",
        dataframe=transactions_df,
        index="id",
        time_index="transaction_time",
    )

    datetime_es = datetime_es.add_dataframe(
        dataframe_name="cards",
        dataframe=cards_df,
        index="id",
    )

    datetime_es = datetime_es.add_relationship("cards", "id", "transactions", "card_id")
    datetime_es.add_last_time_indexes()
    return datetime_es


def test_dfs_empty_features():
    error_text = "No features can be generated from the specified primitives. Please make sure the primitives you are using are compatible with the variable types in your data."
    teams = pd.DataFrame({"id": range(3), "name": ["Breakers", "Spirit", "Thorns"]})
    games = pd.DataFrame(
        {
            "id": range(5),
            "home_team_id": [2, 2, 1, 0, 1],
            "away_team_id": [1, 0, 2, 1, 0],
            "home_team_score": [3, 0, 1, 0, 4],
            "away_team_score": [2, 1, 2, 0, 0],
        },
    )
    dataframes = {
        "teams": (teams, "id", None, {"name": "natural_language"}),
        "games": (games, "id"),
    }
    relationships = [("teams", "id", "games", "home_team_id")]
    with patch.object(DeepFeatureSynthesis, "build_features", return_value=[]):
        features = dfs(
            dataframes,
            relationships,
            target_dataframe_name="teams",
            features_only=True,
        )
        assert features == []
    with pytest.raises(AssertionError, match=error_text), patch.object(
        DeepFeatureSynthesis,
        "build_features",
        return_value=[],
    ):
        dfs(
            dataframes,
            relationships,
            target_dataframe_name="teams",
            features_only=False,
        )


def test_passing_strings_to_logical_types_dfs():
    teams = pd.DataFrame({"id": range(3), "name": ["Breakers", "Spirit", "Thorns"]})
    games = pd.DataFrame(
        {
            "id": range(5),
            "home_team_id": [2, 2, 1, 0, 1],
            "away_team_id": [1, 0, 2, 1, 0],
            "home_team_score": [3, 0, 1, 0, 4],
            "away_team_score": [2, 1, 2, 0, 0],
        },
    )
    dataframes = {
        "teams": (teams, "id", None, {"name": "natural_language"}),
        "games": (games, "id"),
    }
    relationships = [("teams", "id", "games", "home_team_id")]

    features = dfs(
        dataframes,
        relationships,
        target_dataframe_name="teams",
        features_only=True,
    )

    name_logical_type = features[0].dataframe["name"].ww.logical_type
    assert isinstance(name_logical_type, NaturalLanguage)


def test_accepts_cutoff_time_df(dataframes, relationships):
    cutoff_times_df = pd.DataFrame({"instance_id": [1, 2, 3], "time": [10, 12, 15]})
    feature_matrix, features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
    )
    feature_matrix = to_pandas(feature_matrix, index="id", sort_index=True)
    assert len(feature_matrix.index) == 3
    assert len(feature_matrix.columns) == len(features)


@pytest.mark.skipif("not dd")
def test_warns_cutoff_time_dask(dataframes, relationships):
    cutoff_times_df = pd.DataFrame({"instance_id": [1, 2, 3], "time": [10, 12, 15]})
    cutoff_times_df = dd.from_pandas(cutoff_times_df, npartitions=2)
    match = (
        "cutoff_time should be a Pandas DataFrame: "
        "computing cutoff_time, this may take a while"
    )
    with pytest.warns(UserWarning, match=match):
        dfs(
            dataframes=dataframes,
            relationships=relationships,
            target_dataframe_name="transactions",
            cutoff_time=cutoff_times_df,
        )


def test_accepts_cutoff_time_compose(dataframes, relationships):
    def fraud_occured(df):
        return df["fraud"].any()

    kwargs = {
        "time_index": "transaction_time",
        "labeling_function": fraud_occured,
        "window_size": 1,
    }
    if parse(cp.__version__) >= parse("0.10.0"):
        kwargs["target_dataframe_index"] = "card_id"
    else:
        kwargs["target_dataframe_name"] = "card_id"  # pragma: no cover

    lm = cp.LabelMaker(**kwargs)

    transactions_df = to_pandas(dataframes["transactions"][0])

    labels = lm.search(transactions_df, num_examples_per_instance=-1)

    labels["time"] = pd.to_numeric(labels["time"])
    labels.rename({"card_id": "id"}, axis=1, inplace=True)

    feature_matrix, features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="cards",
        cutoff_time=labels,
    )
    feature_matrix = to_pandas(feature_matrix, index="id")
    assert len(feature_matrix.index) == 6
    assert len(feature_matrix.columns) == len(features) + 1


def test_accepts_single_cutoff_time(dataframes, relationships):
    feature_matrix, features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=20,
    )
    feature_matrix = to_pandas(feature_matrix, index="id")
    assert len(feature_matrix.index) == 5
    assert len(feature_matrix.columns) == len(features)


def test_accepts_no_cutoff_time(dataframes, relationships):
    feature_matrix, features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        instance_ids=[1, 2, 3, 5, 6],
    )
    feature_matrix = to_pandas(feature_matrix, index="id")
    assert len(feature_matrix.index) == 5
    assert len(feature_matrix.columns) == len(features)


def test_ignores_instance_ids_if_cutoff_df(dataframes, relationships):
    cutoff_times_df = pd.DataFrame({"instance_id": [1, 2, 3], "time": [10, 12, 15]})
    instance_ids = [1, 2, 3, 4, 5]
    feature_matrix, features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
        instance_ids=instance_ids,
    )
    feature_matrix = to_pandas(feature_matrix, index="id")
    assert len(feature_matrix.index) == 3
    assert len(feature_matrix.columns) == len(features)


def test_approximate_features(pd_dataframes, relationships):
    # TODO: Update to use Dask dataframes when issue #985 is closed
    cutoff_times_df = pd.DataFrame(
        {"instance_id": [1, 3, 1, 5, 3, 6], "time": [11, 16, 16, 26, 17, 22]},
    )
    # force column to BooleanNullable
    pd_dataframes["transactions"] += ({"fraud": "BooleanNullable"},)
    feature_matrix, features = dfs(
        dataframes=pd_dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
        approximate=5,
        cutoff_time_in_index=True,
    )
    direct_agg_feat_name = "cards.PERCENT_TRUE(transactions.fraud)"
    assert len(feature_matrix.index) == 6
    assert len(feature_matrix.columns) == len(features)

    truth_values = pd.Series(data=[1.0, 0.5, 0.5, 1.0, 0.5, 1.0])

    assert (feature_matrix[direct_agg_feat_name] == truth_values.values).all()


def test_all_columns(pd_dataframes, relationships):
    cutoff_times_df = pd.DataFrame({"instance_id": [1, 2, 3], "time": [10, 12, 15]})
    feature_matrix, features = dfs(
        dataframes=pd_dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
        agg_primitives=[Max, Mean, Min, Sum],
        trans_primitives=[],
        groupby_trans_primitives=["cum_sum"],
        max_depth=3,
        allowed_paths=None,
        ignore_dataframes=None,
        ignore_columns=None,
        seed_features=None,
    )
    assert len(feature_matrix.index) == 3
    assert len(feature_matrix.columns) == len(features)


def test_features_only(dataframes, relationships):
    if len(dataframes["transactions"]) > 3:
        dataframes["transactions"][3]["fraud"] = "BooleanNullable"
    else:
        dataframes["transactions"] += ({"fraud": "BooleanNullable"},)
    features = dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        features_only=True,
    )

    # pandas creates 11 features
    # dask creates 10 features (no skew)
    # spark creates 9 features (no skew, no percent_true)
    if isinstance(dataframes["transactions"][0], pd.DataFrame):
        expected_features = 11
    elif is_instance(dataframes["transactions"][0], dd, "DataFrame"):
        expected_features = 10
    else:
        expected_features = 9
    assert len(features) == expected_features


def test_accepts_relative_training_window(datetime_es):
    # TODO: Update to use Dask dataframes when issue #882 is closed
    feature_matrix, _ = dfs(entityset=datetime_es, target_dataframe_name="transactions")

    feature_matrix_2, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-4-1 04:00"),
    )

    feature_matrix_3, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-4-1 04:00"),
        training_window=Timedelta("3 months"),
    )

    feature_matrix_4, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-4-1 04:00"),
        training_window="3 months",
    )

    assert (feature_matrix.index == [1, 2, 3, 4, 5]).all()
    assert (feature_matrix_2.index == [1, 2, 3, 4]).all()
    assert (feature_matrix_3.index == [2, 3, 4]).all()
    assert (feature_matrix_4.index == [2, 3, 4]).all()

    # Test case for leap years
    feature_matrix_5, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-2-29 04:00"),
        training_window=Timedelta("1 year"),
        include_cutoff_time=True,
    )
    assert (feature_matrix_5.index == [2]).all()

    feature_matrix_5, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-2-29 04:00"),
        training_window=Timedelta("1 year"),
        include_cutoff_time=False,
    )
    assert (feature_matrix_5.index == [1, 2]).all()


def test_accepts_pd_timedelta_training_window(datetime_es):
    # TODO: Update to use Dask dataframes when issue #882 is closed
    feature_matrix, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-3-31 04:00"),
        training_window=pd.Timedelta(61, "D"),
    )

    assert (feature_matrix.index == [2, 3, 4]).all()


def test_accepts_pd_dateoffset_training_window(datetime_es):
    # TODO: Update to use Dask dataframes when issue #882 is closed
    feature_matrix, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-3-31 04:00"),
        training_window=pd.DateOffset(months=2),
    )

    feature_matrix_2, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.Timestamp("2012-3-31 04:00"),
        training_window=pd.offsets.BDay(44),
    )

    assert (feature_matrix.index == [2, 3, 4]).all()
    assert (feature_matrix.index == feature_matrix_2.index).all()


def test_accepts_datetime_and_string_offset(datetime_es):
    feature_matrix, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time=pd.to_datetime("2012-3-31 04:00"),
        training_window=pd.DateOffset(months=2),
    )

    feature_matrix_2, _ = dfs(
        entityset=datetime_es,
        target_dataframe_name="transactions",
        cutoff_time="2012-3-31 04:00",
        training_window=pd.offsets.BDay(44),
    )

    assert (feature_matrix.index == [2, 3, 4]).all()
    assert (feature_matrix.index == feature_matrix_2.index).all()


def test_handles_pandas_parser_error(datetime_es):
    with pytest.raises(ValueError):
        _, _ = dfs(
            entityset=datetime_es,
            target_dataframe_name="transactions",
            cutoff_time="2--012-----3-----31 04:00",
            training_window=pd.DateOffset(months=2),
        )


def test_handles_pandas_overflow_error(datetime_es):
    # pandas 1.5.0 raises ValueError, older versions raised OverflowError
    with pytest.raises((OverflowError, ValueError)):
        _, _ = dfs(
            entityset=datetime_es,
            target_dataframe_name="transactions",
            cutoff_time="200000000000000000000000000000000000000000000000000000000000000000-3-31 04:00",
            training_window=pd.DateOffset(months=2),
        )


def test_warns_with_unused_primitives(es):
    if es.dataframe_type == Library.SPARK:
        pytest.skip("Spark throws extra warnings")
    trans_primitives = ["num_characters", "num_words", "add_numeric"]
    agg_primitives = [Max, "min"]

    warning_text = (
        "Some specified primitives were not used during DFS:\n"
        + "  trans_primitives: ['add_numeric']\n  agg_primitives: ['max', 'min']\n"
        + "This may be caused by a using a value of max_depth that is too small, not setting interesting values, "
        + "or it may indicate no compatible columns for the primitive were found in the data. If the DFS call "
        + "contained multiple instances of a primitive in the list above, none of them were used."
    )

    with pytest.warns(UnusedPrimitiveWarning) as record:
        dfs(
            entityset=es,
            target_dataframe_name="customers",
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            max_depth=1,
            features_only=True,
        )

    assert record[0].message.args[0] == warning_text

    # Should not raise a warning
    with pytest.warns(None) as record:
        dfs(
            entityset=es,
            target_dataframe_name="customers",
            trans_primitives=trans_primitives,
            agg_primitives=agg_primitives,
            max_depth=2,
            features_only=True,
        )

    assert not record


def test_no_warns_with_camel_and_title_case(es):
    for trans_primitive in ["isNull", "IsNull"]:
        # Should not raise a UnusedPrimitiveWarning warning
        with pytest.warns(None) as record:
            dfs(
                entityset=es,
                target_dataframe_name="customers",
                trans_primitives=[trans_primitive],
                max_depth=1,
                features_only=True,
            )

        assert not record

    for agg_primitive in ["numUnique", "NumUnique"]:
        # Should not raise a UnusedPrimitiveWarning warning
        with pytest.warns(None) as record:
            dfs(
                entityset=es,
                target_dataframe_name="customers",
                agg_primitives=[agg_primitive],
                max_depth=2,
                features_only=True,
            )

        assert not record


def test_does_not_warn_with_stacking_feature(pd_es):
    with pytest.warns(None) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="régions",
            agg_primitives=["percent_true"],
            trans_primitives=[GreaterThanScalar(5)],
            primitive_options={
                "greater_than_scalar": {"include_dataframes": ["stores"]},
            },
            features_only=True,
        )

    assert not record


def test_warns_with_unused_where_primitives(es):
    if es.dataframe_type == Library.SPARK:
        pytest.skip("Spark throws extra warnings")
    warning_text = (
        "Some specified primitives were not used during DFS:\n"
        + "  where_primitives: ['count', 'sum']\n"
        + "This may be caused by a using a value of max_depth that is too small, not setting interesting values, "
        + "or it may indicate no compatible columns for the primitive were found in the data. If the DFS call "
        + "contained multiple instances of a primitive in the list above, none of them were used."
    )

    with pytest.warns(UnusedPrimitiveWarning) as record:
        dfs(
            entityset=es,
            target_dataframe_name="customers",
            agg_primitives=["count"],
            where_primitives=["sum", "count"],
            max_depth=1,
            features_only=True,
        )

    assert record[0].message.args[0] == warning_text


def test_warns_with_unused_groupby_primitives(pd_es):
    warning_text = (
        "Some specified primitives were not used during DFS:\n"
        + "  groupby_trans_primitives: ['cum_sum']\n"
        + "This may be caused by a using a value of max_depth that is too small, not setting interesting values, "
        + "or it may indicate no compatible columns for the primitive were found in the data. If the DFS call "
        + "contained multiple instances of a primitive in the list above, none of them were used."
    )

    with pytest.warns(UnusedPrimitiveWarning) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="sessions",
            groupby_trans_primitives=["cum_sum"],
            max_depth=1,
            features_only=True,
        )

    assert record[0].message.args[0] == warning_text

    # Should not raise a warning
    with pytest.warns(None) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="customers",
            groupby_trans_primitives=["cum_sum"],
            max_depth=1,
            features_only=True,
        )

    assert not record


def test_warns_with_unused_custom_primitives(pd_es):
    class AboveTen(TransformPrimitive):
        name = "above_ten"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

    trans_primitives = [AboveTen]

    warning_text = (
        "Some specified primitives were not used during DFS:\n"
        + "  trans_primitives: ['above_ten']\n"
        + "This may be caused by a using a value of max_depth that is too small, not setting interesting values, "
        + "or it may indicate no compatible columns for the primitive were found in the data. If the DFS call "
        + "contained multiple instances of a primitive in the list above, none of them were used."
    )

    with pytest.warns(UnusedPrimitiveWarning) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="sessions",
            trans_primitives=trans_primitives,
            max_depth=1,
            features_only=True,
        )

    assert record[0].message.args[0] == warning_text

    # Should not raise a warning
    with pytest.warns(None) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="customers",
            trans_primitives=trans_primitives,
            max_depth=1,
            features_only=True,
        )

    class MaxAboveTen(AggregationPrimitive):
        name = "max_above_ten"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

    agg_primitives = [MaxAboveTen]

    warning_text = (
        "Some specified primitives were not used during DFS:\n"
        + "  agg_primitives: ['max_above_ten']\n"
        + "This may be caused by a using a value of max_depth that is too small, not setting interesting values, "
        + "or it may indicate no compatible columns for the primitive were found in the data. If the DFS call "
        + "contained multiple instances of a primitive in the list above, none of them were used."
    )

    with pytest.warns(UnusedPrimitiveWarning) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="stores",
            agg_primitives=agg_primitives,
            max_depth=1,
            features_only=True,
        )

    assert record[0].message.args[0] == warning_text

    # Should not raise a warning
    with pytest.warns(None) as record:
        dfs(
            entityset=pd_es,
            target_dataframe_name="sessions",
            agg_primitives=agg_primitives,
            max_depth=1,
            features_only=True,
        )


def test_calls_progress_callback(dataframes, relationships):
    class MockProgressCallback:
        def __init__(self):
            self.progress_history = []
            self.total_update = 0
            self.total_progress_percent = 0

        def __call__(self, update, progress_percent, time_elapsed):
            self.total_update += update
            self.total_progress_percent = progress_percent
            self.progress_history.append(progress_percent)

    mock_progress_callback = MockProgressCallback()

    dfs(
        dataframes=dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        progress_callback=mock_progress_callback,
    )

    # second to last entry is the last update from feature calculation
    assert np.isclose(
        mock_progress_callback.progress_history[-2],
        FEATURE_CALCULATION_PERCENTAGE * 100,
    )
    assert np.isclose(mock_progress_callback.total_update, 100.0)
    assert np.isclose(mock_progress_callback.total_progress_percent, 100.0)


def test_calls_progress_callback_cluster(pd_dataframes, relationships, dask_cluster):
    class MockProgressCallback:
        def __init__(self):
            self.progress_history = []
            self.total_update = 0
            self.total_progress_percent = 0

        def __call__(self, update, progress_percent, time_elapsed):
            self.total_update += update
            self.total_progress_percent = progress_percent
            self.progress_history.append(progress_percent)

    mock_progress_callback = MockProgressCallback()

    dkwargs = {"cluster": dask_cluster.scheduler.address}
    dfs(
        dataframes=pd_dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        progress_callback=mock_progress_callback,
        dask_kwargs=dkwargs,
    )

    assert np.isclose(mock_progress_callback.total_update, 100.0)
    assert np.isclose(mock_progress_callback.total_progress_percent, 100.0)


def test_dask_kwargs(pd_dataframes, relationships, dask_cluster):
    cutoff_times_df = pd.DataFrame({"instance_id": [1, 2, 3], "time": [10, 12, 15]})
    feature_matrix, features = dfs(
        dataframes=pd_dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
    )

    dask_kwargs = {"cluster": dask_cluster.scheduler.address}
    feature_matrix_2, features_2 = dfs(
        dataframes=pd_dataframes,
        relationships=relationships,
        target_dataframe_name="transactions",
        cutoff_time=cutoff_times_df,
        dask_kwargs=dask_kwargs,
    )

    assert all(
        f1.unique_name() == f2.unique_name() for f1, f2 in zip(features, features_2)
    )
    for column in feature_matrix:
        for x, y in zip(feature_matrix[column], feature_matrix_2[column]):
            assert (pd.isnull(x) and pd.isnull(y)) or (x == y)



import re
from string import punctuation

import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive
from featuretools.primitives.standard.transform.natural_language.constants import (
    DELIMITERS,
)


class UpperCaseWordCount(TransformPrimitive):
    """Determines the number of words in a string that are entirely capitalized.

    Description:
        Given list of strings, determine the number of words in each string
        that are entirely capitalized.

        If a string is missing, return `NaN`.

    Examples:
        >>> x = ['This IS a string.', 'This is a string', 'AAA']
        >>> upper_case_word_count = UpperCaseWordCount()
        >>> upper_case_word_count(x).tolist()
        [1, 0, 1]
    """

    name = "upper_case_word_count"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    default_value = 0

    def get_function(self):
        def upper_case_word_count(x):
            def _count_upper_case_words(elem):
                if pd.isna(elem):
                    return pd.NA
                return sum(
                    1
                    for word in re.split(DELIMITERS, elem)
                    if word.strip(punctuation) and word.upper() == word
                )

            return x.apply(_count_upper_case_words)

        return upper_case_word_count



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class Not(TransformPrimitive):
    """Negates a boolean value.

    Examples:
        >>> not_func = Not()
        >>> not_func([True, True, False]).tolist()
        [False, False, True]
    """

    name = "not"
    input_types = [
        [ColumnSchema(logical_type=Boolean)],
        [ColumnSchema(logical_type=BooleanNullable)],
    ]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the negation of {}"

    def generate_name(self, base_feature_names):
        return "NOT({})".format(base_feature_names[0])

    def get_function(self):
        return np.logical_not



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class First(AggregationPrimitive):
    """Determines the first value in a list.

    Examples:
        >>> first = First()
        >>> first([1, 2, 3, 4, 5, None])
        1.0
    """

    name = "first"
    input_types = [ColumnSchema()]
    return_type = None
    stack_on_self = False
    description_template = "the first instance of {}"

    def get_function(self, agg_type=Library.PANDAS):
        def pd_first(x):
            return x.iloc[0]

        return pd_first



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Integer

from featuretools.primitives.base import AggregationPrimitive


class NUniqueWeeks(AggregationPrimitive):
    """Determines the number of unique weeks.

    Description:
        Given a list of datetimes, return the number of unique
        weeks (Monday-Sunday). NUniqueWeeks counts by absolute
        week, not week of year, so the first week of 2018 and
        the first week of 2019 count as two unique values.

    Examples:
        >>> from datetime import datetime
        >>> n_unique_weeks = NUniqueWeeks()
        >>> times = [datetime(2018, 2, 2),
        ...          datetime(2019, 1, 1),
        ...          datetime(2019, 2, 1),
        ...          datetime(2019, 2, 1),
        ...          datetime(2019, 2, 3),
        ...          datetime(2019, 2, 21)]
        >>> n_unique_weeks(times)
        4
    """

    name = "n_unique_weeks"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def n_unique_weeks(x):
            return x.dt.to_period("W").nunique()

        return n_unique_weeks



import pytest
from woodwork.column_schema import ColumnSchema

from featuretools.primitives import (
    AggregationPrimitive,
    Count,
    Hour,
    IsIn,
    Not,
    TimeSincePrevious,
    TransformPrimitive,
)
from featuretools.synthesis.get_valid_primitives import get_valid_primitives
from featuretools.utils.gen_utils import Library


def test_get_valid_primitives_selected_primitives(es):
    agg_prims, trans_prims = get_valid_primitives(
        es,
        "log",
        selected_primitives=[Hour, Count],
    )
    assert set(agg_prims) == set([Count])
    assert set(trans_prims) == set([Hour])

    agg_prims, trans_prims = get_valid_primitives(
        es,
        "products",
        selected_primitives=[Hour],
        max_depth=1,
    )
    assert set(agg_prims) == set()
    assert set(trans_prims) == set()


def test_get_valid_primitives_selected_primitives_strings(es):
    agg_prims, trans_prims = get_valid_primitives(
        es,
        "log",
        selected_primitives=["hour", "count"],
    )
    assert set(agg_prims) == set([Count])
    assert set(trans_prims) == set([Hour])

    agg_prims, trans_prims = get_valid_primitives(
        es,
        "products",
        selected_primitives=["hour"],
        max_depth=1,
    )
    assert set(agg_prims) == set()
    assert set(trans_prims) == set()


def test_invalid_primitive(es):
    with pytest.raises(ValueError, match="'foobar' is not a recognized primitive name"):
        get_valid_primitives(
            es,
            target_dataframe_name="log",
            selected_primitives=["foobar"],
        )

    msg = (
        "Selected primitive <enum 'Library'> "
        "is not an AggregationPrimitive, TransformPrimitive, or str"
    )
    with pytest.raises(ValueError, match=msg):
        get_valid_primitives(
            es,
            target_dataframe_name="log",
            selected_primitives=[Library],
        )


def test_primitive_compatibility(es):
    _, trans_prims = get_valid_primitives(
        es,
        "customers",
        selected_primitives=[TimeSincePrevious],
    )

    if es.dataframe_type != Library.PANDAS:
        assert len(trans_prims) == 0
    else:
        assert len(trans_prims) == 1


def test_get_valid_primitives_custom_primitives(pd_es):
    class ThreeMostCommonCat(AggregationPrimitive):
        name = "n_most_common_categorical"
        input_types = [ColumnSchema(semantic_tags={"category"})]
        return_type = ColumnSchema(semantic_tags={"category"})
        number_output_features = 3

    class AddThree(TransformPrimitive):
        name = "add_three"
        input_types = [
            ColumnSchema(semantic_tags="numeric"),
            ColumnSchema(semantic_tags="numeric"),
            ColumnSchema(semantic_tags="numeric"),
        ]
        return_type = ColumnSchema(semantic_tags="numeric")
        commutative = True
        compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]

    agg_prims, trans_prims = get_valid_primitives(pd_es, "log")
    assert ThreeMostCommonCat not in agg_prims
    assert AddThree not in trans_prims

    with pytest.raises(
        ValueError,
        match="'add_three' is not a recognized primitive name",
    ):
        agg_prims, trans_prims = get_valid_primitives(
            pd_es,
            "log",
            2,
            [ThreeMostCommonCat, "add_three"],
        )


def test_get_valid_primitives_all_primitives(es):
    agg_prims, trans_prims = get_valid_primitives(es, "customers")
    assert Count in agg_prims
    assert Hour in trans_prims


def test_get_valid_primitives_single_table(transform_es):
    msg = "Only one dataframe in entityset, changing max_depth to 1 since deeper features cannot be created"
    with pytest.warns(UserWarning, match=msg):
        agg_prims, trans_prims = get_valid_primitives(transform_es, "first")

    assert set(agg_prims) == set()
    assert IsIn in trans_prims


def test_get_valid_primitives_with_dfs_kwargs(es):
    agg_prims, trans_prims = get_valid_primitives(
        es,
        "customers",
        selected_primitives=[Hour, Count, Not],
    )
    assert set(agg_prims) == set([Count])
    assert set(trans_prims) == set([Hour, Not])

    # Can use other dfs parameters and they get applied
    agg_prims, trans_prims = get_valid_primitives(
        es,
        "customers",
        selected_primitives=[Hour, Count, Not],
        ignore_columns={"customers": ["loves_ice_cream"]},
    )
    assert set(agg_prims) == set([Count])
    assert set(trans_prims) == set([Hour])

    agg_prims, trans_prims = get_valid_primitives(
        es,
        "products",
        selected_primitives=[Hour, Count],
        ignore_dataframes=["log"],
    )
    assert set(agg_prims) == set()
    assert set(trans_prims) == set()



from typing import Callable, Optional, Union

import numpy as np
import pandas as pd
from pandas import Series
from pandas.core.window.rolling import Rolling
from pandas.tseries.frequencies import to_offset


def roll_series_with_gap(
    series: Series,
    window_length: Union[int, str],
    gap: Union[int, str],
    min_periods: int,
) -> Rolling:
    """Provide rolling window calculations where the windows are determined using both a gap parameter
    that indicates the amount of time between each instance and its window and a window length parameter
    that determines the amount of data in each window.

    Args:
        series (Series): The series over which rolling windows will be created. The series must have numeric values and a DatetimeIndex.
        window_length (int, string): Specifies the amount of data included in each window.
            If an integer is provided, it will correspond to a number of rows. For data with a uniform sampling frequency,
            for example of one day, the window_length will correspond to a period of time, in this case,
            7 days for a window_length of 7.
            If a string is provided, it must be one of pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time that each window should span.
            The list of available offset aliases can be found at
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        gap (int, string, optional): Specifies a gap backwards from each instance before the
            window of usable data begins. If an integer is provided, it will correspond to a number of rows.
            If a string is provided, it must be one of pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time between a target instance and the beginning of its window.
            Defaults to 0, which will include the target instance in the window.
        min_periods (int, optional): Minimum number of observations required for performing calculations
            over the window. Can only be as large as window_length when window_length is an integer.
            When window_length is an offset alias string, this limitation does not exist, but care should be taken
            to not choose a min_periods that will always be larger than the number of observations in a window.
            Defaults to 1.

    Returns:
        pandas.core.window.rolling.Rolling: The Rolling object for the series passed in.
    """
    _check_window_length(window_length)
    _check_gap(window_length, gap)

    functional_window_length = window_length
    if isinstance(gap, str):
        # Add the window_length and gap so that the rolling operation correctly takes gap into account.
        # That way, we can later remove the gap rows in order to apply the primitive function
        # to the correct window
        functional_window_length = to_offset(window_length) + to_offset(gap)
    elif gap > 0:
        # When gap is numeric, we can apply a shift to incorporate gap right now
        # since the gap will be the same number of rows for the whole dataset
        series = series.shift(gap)

    return series.rolling(functional_window_length, min_periods)


def _get_rolled_series_without_gap(window: Series, gap_offset: str) -> Series:
    """Applies the gap offset_string to the rolled window, returning a window
    that is the correct length of time away from the original instance.

    Args:
        window (Series): A rolling window that includes both the window length and gap spans of time.
        gap_offset (string): The pandas offset alias that determines how much time at the end of the window
            should be removed.

    Returns:
        Series: The window with gap rows removed
    """
    if not len(window):
        return window

    window_start_date = window.index[0]
    window_end_date = window.index[-1]

    gap_bound = window_end_date - to_offset(gap_offset)

    # If the gap is larger than the series, no rows are left in the window
    if gap_bound < window_start_date:
        return Series(dtype="float64")

    # Only return the rows that are within the offset's bounds
    return window[window.index <= gap_bound]


def apply_roll_with_offset_gap(
    window: Series,
    gap_offset: str,
    reducer_fn: Callable[[Series], float],
    min_periods: int,
) -> float:
    """Takes in a series to which an offset gap will be applied, removing however many
    rows fall under the gap before applying the reducing function.

    Args:
        window (Series):  A rolling window that includes both the window length and gap spans of time.
        gap_offset (string): The pandas offset alias that determines how much time at the end of the window
            should be removed.
        reducer_fn (callable[Series -> float]): The function to be applied to the window in order to produce
            the aggregate that will be included in the resulting feature.
        min_periods (int): Minimum number of observations required for performing calculations
            over the window.

    Returns:
        float: The aggregate value to be used as a feature value.
    """
    window = _get_rolled_series_without_gap(window, gap_offset)

    if min_periods is None:
        min_periods = 1

    if len(window) < min_periods or not len(window):
        return np.nan

    return reducer_fn(window)


def _check_window_length(window_length: Union[int, str]) -> None:
    # Window length must either be a valid offset alias
    if isinstance(window_length, str):
        try:
            to_offset(window_length)
        except ValueError:
            raise ValueError(
                f"Cannot roll series. The specified window length, {window_length}, is not a valid offset alias.",
            )
    # Or an integer greater than zero
    elif isinstance(window_length, int):
        if window_length <= 0:
            raise ValueError("Window length must be greater than zero.")
    else:
        raise TypeError("Window length must be either an offset string or an integer.")


def _check_gap(window_length: Union[int, str], gap: Union[int, str]) -> None:
    # Gap must either be a valid offset string that also has an offset string window length
    if isinstance(gap, str):
        if not isinstance(window_length, str):
            raise TypeError(
                f"Cannot roll series with offset gap, {gap}, and numeric window length, {window_length}. "
                "If an offset alias is used for gap, the window length must also be defined as an offset alias. "
                "Please either change gap to be numeric or change window length to be an offset alias.",
            )
        try:
            to_offset(gap)
        except ValueError:
            raise ValueError(
                f"Cannot roll series. The specified gap, {gap}, is not a valid offset alias.",
            )
    # Or an integer greater than or equal to zero
    elif isinstance(gap, int):
        if gap < 0:
            raise ValueError("Gap must be greater than or equal to zero.")
    else:
        raise TypeError("Gap must be either an offset string or an integer.")


def apply_rolling_agg_to_series(
    series: Series,
    agg_func: Callable[[Series], float],
    window_length: Union[int, str],
    gap: Union[int, str] = 0,
    min_periods: int = 1,
    ignore_window_nans: bool = False,
) -> np.ndarray:
    """Applies a given aggregation function to a rolled series.

    Args:
        series (Series): The series over which rolling windows will be created. The series must have numeric values and a DatetimeIndex.
        agg_func (callable[Series -> float]): The aggregation function to apply to a rolled series.
        window_length (int, string): Specifies the amount of data included in each window.
            If an integer is provided, it will correspond to a number of rows. For data with a uniform sampling frequency,
            for example of one day, the window_length will correspond to a period of time, in this case,
            7 days for a window_length of 7.
            If a string is provided, it must be one of pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time that each window should span.
            The list of available offset aliases can be found at
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        gap (int, string, optional): Specifies a gap backwards from each instance before the
            window of usable data begins. If an integer is provided, it will correspond to a number of rows.
            If a string is provided, it must be one of pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time between a target instance and the beginning of its window.
            Defaults to 0, which will include the target instance in the window.
        min_periods (int, optional): Minimum number of observations required for performing calculations
            over the window. Can only be as large as window_length when window_length is an integer.
            When window_length is an offset alias string, this limitation does not exist, but care should be taken
            to not choose a min_periods that will always be larger than the number of observations in a window.
            Defaults to 1.
        ignore_window_nans (bool, optional): Whether or not NaNs in the rolling window should be included in the rolling calculation.
            NaNs by default get counted towards min_periods. When set to True,
            all partial values calculated by `agg_func` in the rolling window get replaced with NaN.
            Defaults to False.

    Returns:
        numpy.ndarray: The array of rolling calculated values.

    Note:
        Certain operations, like `pandas.core.window.rolling.Rolling.count` that can be performed
        on the Rolling object returned here may treat NaNs as periods to include in window calculations.
        So a window [NaN, 1, 3]  when `min_periods=3` will proceed with count, saying there are three periods
        but only two values and would return count=2. The calculation `max` on the other hand,
        would not recognize NaN as a valid period, and would therefore return `max=NaN` as the window has
        less valid periods (two, in this case) than `min_periods` (three, in this case).
        Most rolling calculations act this way. The implication of that here is that in order to
        achieve the gap, we insert NaNs at the beginning of the series, which would cause `count` to calculate
        on windows that technically should not have the correct number of periods. Any primitive that uses this function
        should determine whether `ignore_window_nans` should be set to `true`.

    Note:
        Only offset aliases with fixed frequencies can be used when defining gap and window_length.
        This means that aliases such as `M` or `W` cannot be used, as they can indicate different
        numbers of days. ('M', because different months have different numbers of days;
        'W' because week will indicate a certain day of the week, like W-Wed, so that will
        indicate a different number of days depending on the anchoring date.)

    Note:
        When using an offset alias to define `gap`, an offset alias must also be used to define `window_length`.
        This limitation does not exist when using an offset alias to define `window_length`. In fact,
        if the data has a uniform sampling frequency, it is preferable to use a numeric `gap` as it is more
        efficient."""
    rolled_series = roll_series_with_gap(series, window_length, gap, min_periods)
    if isinstance(gap, str):
        additional_args = (gap, agg_func, min_periods)
        return rolled_series.apply(
            apply_roll_with_offset_gap,
            args=additional_args,
        ).values
    applied_rolled_series = rolled_series.apply(agg_func)

    if ignore_window_nans:
        if not min_periods:
            # when min periods is 0 or None it's treated the same as if it's 1
            num_nans = gap
        else:
            num_nans = min_periods - 1 + gap
        applied_rolled_series.iloc[range(num_nans)] = np.nan
    return applied_rolled_series.values


def _apply_gap_for_expanding_primitives(
    x: Union[Series, pd.Index],
    gap: Union[int, str],
) -> Optional[Series]:
    if not isinstance(gap, int):
        raise TypeError(
            "String offsets are not supported for the gap parameter in Expanding primitives",
        )
    if isinstance(x, pd.Index):
        return x.to_series().shift(gap)
    return x.shift(gap)



import pandas as pd
from woodwork.logical_types import Boolean, BooleanNullable


def remove_low_information_features(feature_matrix, features=None):
    """Select features that have at least 2 unique values and that are not all null

    Args:
        feature_matrix (:class:`pd.DataFrame`): DataFrame whose columns are feature names and rows are instances
        features (list[:class:`featuretools.FeatureBase`] or list[str], optional): List of features to select

    Returns:
        (feature_matrix, features)

    """
    keep = [
        c
        for c in feature_matrix
        if (
            feature_matrix[c].nunique(dropna=False) > 1
            and feature_matrix[c].dropna().shape[0] > 0
        )
    ]
    feature_matrix = feature_matrix[keep]
    if features is not None:
        features = [f for f in features if f.get_name() in feature_matrix.columns]
        return feature_matrix, features
    return feature_matrix


def remove_highly_null_features(feature_matrix, features=None, pct_null_threshold=0.95):
    """
    Removes columns from a feature matrix that have higher than a set threshold
    of null values.

    Args:
        feature_matrix (:class:`pd.DataFrame`): DataFrame whose columns are feature names and rows are instances.
        features (list[:class:`featuretools.FeatureBase`] or list[str], optional): List of features to select.
        pct_null_threshold (float): If the percentage of NaN values in an input feature exceeds this amount,
                that feature will be considered highly-null. Defaults to 0.95.

    Returns:
        pd.DataFrame, list[:class:`.FeatureBase`]:
            The feature matrix and the list of generated feature definitions. Matches dfs output.
            If no feature list is provided as input, the feature list will not be returned.
    """
    if pct_null_threshold < 0 or pct_null_threshold > 1:
        raise ValueError(
            "pct_null_threshold must be a float between 0 and 1, inclusive.",
        )

    percent_null_by_col = (feature_matrix.isnull().mean()).to_dict()

    if pct_null_threshold == 0.0:
        keep = [
            f_name
            for f_name, pct_null in percent_null_by_col.items()
            if pct_null <= pct_null_threshold
        ]
    else:
        keep = [
            f_name
            for f_name, pct_null in percent_null_by_col.items()
            if pct_null < pct_null_threshold
        ]

    return _apply_feature_selection(keep, feature_matrix, features)


def remove_single_value_features(
    feature_matrix,
    features=None,
    count_nan_as_value=False,
):
    """Removes columns in feature matrix where all the values are the same.

    Args:
        feature_matrix (:class:`pd.DataFrame`): DataFrame whose columns are feature names and rows are instances.
        features (list[:class:`featuretools.FeatureBase`] or list[str], optional): List of features to select.
        count_nan_as_value (bool): If True, missing values will be counted as their own unique value.
                    If set to False, a feature that has one unique value and all other
                    data missing will be removed from the feature matrix. Defaults to False.

     Returns:
        pd.DataFrame, list[:class:`.FeatureBase`]:
            The feature matrix and the list of generated feature definitions.
            Matches dfs output.
            If no feature list is provided as input, the feature list will not be returned.
    """
    unique_counts_by_col = feature_matrix.nunique(
        dropna=not count_nan_as_value,
    ).to_dict()

    keep = [
        f_name
        for f_name, unique_count in unique_counts_by_col.items()
        if unique_count > 1
    ]
    return _apply_feature_selection(keep, feature_matrix, features)


def remove_highly_correlated_features(
    feature_matrix,
    features=None,
    pct_corr_threshold=0.95,
    features_to_check=None,
    features_to_keep=None,
):
    """Removes columns in feature matrix that are highly correlated with another column.

    Note:
        We make the assumption that, for a pair of features, the feature that is further
        right in the feature matrix produced by ``dfs`` is the more complex one.
        The assumption does not hold if the order of columns in the feature
        matrix has changed from what ``dfs`` produces.

    Args:
        feature_matrix (:class:`pd.DataFrame`): DataFrame whose columns are feature
                    names and rows are instances. If Woodwork is not initalized, will
                    perform Woodwork initialization, which may result in slightly different
                    types than those in the original feature matrix created by Featuretools.
        features (list[:class:`featuretools.FeatureBase`] or list[str], optional):
                    List of features to select.
        pct_corr_threshold (float): The correlation threshold to be considered highly
                    correlated. Defaults to 0.95.
        features_to_check (list[str], optional): List of column names to check
                    whether any pairs are highly correlated. Will not check any
                    other columns, meaning the only columns that can be removed
                    are in this list. If null, defaults to checking all columns.
        features_to_keep (list[str], optional): List of colum names to keep even
                    if correlated to another column. If null, all columns will be
                    candidates for removal.

    Returns:
        pd.DataFrame, list[:class:`.FeatureBase`]:
            The feature matrix and the list of generated feature definitions.
            Matches dfs output. If no feature list is provided as input,
            the feature list will not be returned. For consistent results,
            do not change the order of features outputted by dfs.
    """
    if feature_matrix.ww.schema is None:
        feature_matrix.ww.init()

    if pct_corr_threshold < 0 or pct_corr_threshold > 1:
        raise ValueError(
            "pct_corr_threshold must be a float between 0 and 1, inclusive.",
        )

    if features_to_check is None:
        features_to_check = list(feature_matrix.columns)
    else:
        for f_name in features_to_check:
            assert (
                f_name in feature_matrix.columns
            ), "feature named {} is not in feature matrix".format(f_name)

    if features_to_keep is None:
        features_to_keep = []

    to_select = ["numeric", Boolean, BooleanNullable]
    fm = feature_matrix.ww[features_to_check]
    fm_to_check = fm.ww.select(include=to_select)

    dropped = set()
    columns_to_check = fm_to_check.columns
    # When two features are found to be highly correlated,
    # we drop the more complex feature
    # Columns produced later in dfs are more complex
    for i in range(len(columns_to_check) - 1, 0, -1):
        more_complex_name = columns_to_check[i]
        more_complex_col = fm_to_check[more_complex_name]

        # Convert boolean or Int64 column to be float64
        if pd.api.types.is_bool_dtype(more_complex_col) or isinstance(
            more_complex_col.dtype,
            pd.Int64Dtype,
        ):
            more_complex_col = more_complex_col.astype("float64")

        for j in range(i - 1, -1, -1):
            less_complex_name = columns_to_check[j]
            less_complex_col = fm_to_check[less_complex_name]

            # Convert boolean or Int64 column to be float64
            if pd.api.types.is_bool_dtype(less_complex_col) or isinstance(
                less_complex_col.dtype,
                pd.Int64Dtype,
            ):
                less_complex_col = less_complex_col.astype("float64")

            if abs(more_complex_col.corr(less_complex_col)) >= pct_corr_threshold:
                dropped.add(more_complex_name)
                break

    keep = [
        f_name
        for f_name in feature_matrix.columns
        if (f_name in features_to_keep or f_name not in dropped)
    ]
    return _apply_feature_selection(keep, feature_matrix, features)


def _apply_feature_selection(keep, feature_matrix, features=None):
    new_matrix = feature_matrix[keep]
    new_feature_names = set(new_matrix.columns)

    if features is not None:
        new_features = []
        for f in features:
            if f.number_output_features > 1:
                slices = [
                    f[i]
                    for i in range(f.number_output_features)
                    if f[i].get_name() in new_feature_names
                ]
                if len(slices) == f.number_output_features:
                    new_features.append(f)
                else:
                    new_features.extend(slices)
            else:
                if f.get_name() in new_feature_names:
                    new_features.append(f)

        return new_matrix, new_features
    return new_matrix



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, Datetime

from featuretools.primitives.base import TransformPrimitive


class IsFirstWeekOfMonth(TransformPrimitive):
    """Determines if a date falls in the first week of the month.

    Description:
        Converts a datetime to a boolean indicating if the date
        falls in the first week of the month. The first week of
        the month starts on day 1, and the week number is incremented
        each Sunday.

    Examples:
        >>> from datetime import datetime
        >>> is_first_week_of_month = IsFirstWeekOfMonth()
        >>> times = [datetime(2019, 3, 1),
        ...          datetime(2019, 3, 3),
        ...          datetime(2019, 3, 31),
        ...          datetime(2019, 3, 30)]
        >>> is_first_week_of_month(times).tolist()
        [True, False, False, False]
    """

    name = "is_first_week_of_month"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=BooleanNullable)

    def get_function(self):
        def is_first_week_of_month(x):
            df = pd.DataFrame({"date": x})
            df["first_day"] = df.date - pd.to_timedelta(df["date"].dt.day - 1, unit="d")
            df["dom"] = df.date.dt.day
            df["first_day_weekday"] = df.first_day.dt.weekday
            df["adjusted_dom"] = df.dom + df.first_day_weekday + 1
            df.loc[df["first_day_weekday"].astype(float) == 6.0, "adjusted_dom"] = df[
                "dom"
            ]
            df["is_first_week"] = np.ceil(df.adjusted_dom / 7.0) == 1.0
            if df["date"].isnull().values.any():
                df["is_first_week"] = df["is_first_week"].astype("object")
                df.loc[df["date"].isnull(), "is_first_week"] = np.nan
            return df.is_first_week.values

        return is_first_week_of_month



from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class DivideNumeric(TransformPrimitive):
    """Performs element-wise division of two lists.

    Description:
        Given a list of values X and a list of values
        Y, determine the quotient of each value in X
        divided by its corresponding value in Y.

    Args:
        commutative (bool): determines if Deep Feature Synthesis should
            generate both x / y and y / x, or just one. If True, there is
            no guarantee which of the two will be generated. Defaults to False.

    Examples:
        >>> divide_numeric = DivideNumeric()
        >>> divide_numeric([2.0, 1.0, 2.0], [1.0, 2.0, 2.0]).tolist()
        [2.0, 0.5, 1.0]
    """

    name = "divide_numeric"
    input_types = [
        ColumnSchema(semantic_tags={"numeric"}),
        ColumnSchema(semantic_tags={"numeric"}),
    ]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the result of {} divided by {}"

    def __init__(self, commutative=False):
        self.commutative = commutative

    def get_function(self):
        def divide_numeric(val1, val2):
            return val1 / val2

        return divide_numeric

    def generate_name(self, base_feature_names):
        return "%s / %s" % (base_feature_names[0], base_feature_names[1])



import numpy as np


def _haversine_calculate(lat_1s, lon_1s, lat_2s, lon_2s, unit):
    # https://stackoverflow.com/a/29546836/2512385
    lon1, lat1, lon2, lat2 = map(np.radians, [lon_1s, lat_1s, lon_2s, lat_2s])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2
    radius_earth = 3958.7613
    if unit == "kilometers":
        radius_earth = 6371.0088
    distances = radius_earth * 2 * np.arcsin(np.sqrt(a))
    return distances



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library, import_or_none

dd = import_or_none("dask.dataframe")


class Any(AggregationPrimitive):
    """Determines if any value is 'True' in a list.

    Description:
        Given a list of booleans, return `True` if one or
        more of the values are `True`.

    Examples:
        >>> any = Any()
        >>> any([False, False, False, True])
        True
    """

    name = "any"
    input_types = [
        [ColumnSchema(logical_type=Boolean)],
        [ColumnSchema(logical_type=BooleanNullable)],
    ]
    return_type = ColumnSchema(logical_type=Boolean)
    stack_on_self = False
    compatibility = [Library.PANDAS, Library.DASK]
    description_template = "whether any of {} are true"

    def get_function(self, agg_type=Library.PANDAS):
        if agg_type == Library.DASK:

            def chunk(s):
                return s.agg(np.any)

            def agg(s):
                return s.agg(np.any)

            return dd.Aggregation(self.name, chunk=chunk, agg=agg)

        return np.any



from featuretools.primitives.base.primitive_base import PrimitiveBase


class AggregationPrimitive(PrimitiveBase):
    def generate_name(
        self,
        base_feature_names,
        relationship_path_name,
        parent_dataframe_name,
        where_str,
        use_prev_str,
    ):
        base_features_str = ", ".join(base_feature_names)
        return "%s(%s.%s%s%s%s)" % (
            self.name.upper(),
            relationship_path_name,
            base_features_str,
            where_str,
            use_prev_str,
            self.get_args_string(),
        )

    def generate_names(
        self,
        base_feature_names,
        relationship_path_name,
        parent_dataframe_name,
        where_str,
        use_prev_str,
    ):
        n = self.number_output_features
        base_name = self.generate_name(
            base_feature_names,
            relationship_path_name,
            parent_dataframe_name,
            where_str,
            use_prev_str,
        )
        return [base_name + "[%s]" % i for i in range(n)]



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class IsNull(TransformPrimitive):
    """Determines if a value is null.

    Examples:
        >>> is_null = IsNull()
        >>> is_null([1, None, 3]).tolist()
        [False, True, False]
    """

    name = "is_null"
    input_types = [ColumnSchema()]
    return_type = ColumnSchema(logical_type=Boolean)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} is null"

    def get_function(self):
        def isnull(array):
            return array.isnull()

        return isnull



# -*- coding: utf-8 -*-
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.standard.transform.natural_language.count_string import (
    CountString,
)


class TitleWordCount(CountString):
    """Determines the number of title words in a string.

    Description:
        Given list of strings, determine the number of title words
        in each string. A title word is defined as any word starting
        with a capital letter. Words at the start of a sentence will
        be counted.

        If a string is missing, return `NaN`.

    Examples:
        >>> x = ['My favorite movie is Jaws.', 'this is a string', 'AAA']
        >>> title_word_count = TitleWordCount()
        >>> title_word_count(x).tolist()
        [2.0, 0.0, 1.0]
    """

    name = "title_word_count"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    default_value = 0

    def __init__(self):
        pattern = r"([A-Z][^\s]*)"
        super().__init__(string=pattern, is_regex=True, ignore_case=False)



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, BooleanNullable, Datetime, Double

from featuretools.primitives.base import AggregationPrimitive


class TimeSinceLastTrue(AggregationPrimitive):
    """Calculates the time since the last `True` value.

    Description:
        Using a series of Datetimes and a series of Booleans, find the last
        record with a `True` value. Return the seconds elapsed between that record
        and the instance's cutoff time. Return nan if no values are `True`.

    Examples:
        >>> from datetime import datetime
        >>> time_since_last_true = TimeSinceLastTrue()
        >>> cutoff_time = datetime(2010, 1, 1, 12, 0, 0)
        >>> times = [datetime(2010, 1, 1, 11, 45, 0),
        ...          datetime(2010, 1, 1, 11, 55, 15),
        ...          datetime(2010, 1, 1, 11, 57, 30)]
        >>> booleans = [True, True, False]
        >>> time_since_last_true(times, booleans, time=cutoff_time)
        285.0
    """

    name = "time_since_last_true"
    input_types = [
        [
            ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
            ColumnSchema(logical_type=Boolean),
        ],
        [
            ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
            ColumnSchema(logical_type=BooleanNullable),
        ],
    ]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    uses_calc_time = True
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def time_since_last_true(datetime_col, bool_col, time=None):
            df = pd.DataFrame(
                {
                    "datetime": datetime_col,
                    "bool": bool_col,
                },
            ).dropna()
            if df.empty:
                return np.nan
            true_indices = df[df["bool"]]
            if true_indices.empty:
                return np.nan
            last_true_index = true_indices.index[-1]
            time_since = time - datetime_col.loc[last_true_index]
            return time_since.total_seconds()

        return time_since_last_true



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Integer

from featuretools.primitives.base import AggregationPrimitive


class NumZeroCrossings(AggregationPrimitive):
    """Determines the number of times a list crosses 0.
    Description:
        Given a list of numbers, return the number of times the value
        crosses 0. It is the number of times the value goes from a
        positive number to a negative number, or a negative number to
        a positive number. NaN values are ignored.
    Examples:
        >>> num_zero_crossings = NumZeroCrossings()
        >>> num_zero_crossings([1, -1, 2, -2, 3, -3])
        5
    """

    name = "num_zero_crossings"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Integer, semantic_tags={"numeric"})

    def get_function(self):
        def num_zero_crossings(x):
            cleaned = x[(x != 0) & (x == x)]
            signs = np.sign(cleaned)
            difference = np.diff(signs)
            crossings = np.where(difference)[0]
            return len(crossings)

        return num_zero_crossings



from featuretools.tests.entry_point_tests.utils import (
    _import_featuretools,
    _install_featuretools_primitives,
    _python,
    _uninstall_featuretools_primitives,
)


def test_entry_point():
    _install_featuretools_primitives()
    featuretools_log = _import_featuretools("debug").stdout.decode()
    new_primitive = _python("-c", "from featuretools.primitives import NewPrimitive")
    _uninstall_featuretools_primitives()
    assert new_primitive.returncode == 0

    invalid_primitive = 'Featuretools failed to load "invalid" primitives from "featuretools_primitives.invalid_primitive". '
    invalid_primitive += "For a full stack trace, set logging to debug."
    assert invalid_primitive in featuretools_log

    existing_primitive = 'While loading primitives via "existing" entry point, '
    existing_primitive += 'ignored primitive "Sum" from "featuretools_primitives.existing_primitive" because a primitive '
    existing_primitive += 'with that name already exists in "featuretools.primitives.standard.aggregation.sum_primitive"'
    assert existing_primitive in featuretools_log



# flake8: noqa
from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.primitives.standard.aggregation import *
from featuretools.primitives.standard.transform import *



import pandas as pd
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class Skew(AggregationPrimitive):
    """Computes the extent to which a distribution differs from a normal distribution.

    Description:
        For normally distributed data, the skewness should be about 0.
        A skewness value > 0 means that there is more weight in the
        left tail of the distribution.

    Examples:
        >>> skew = Skew()
        >>> skew([1, 10, 30, None])
        1.0437603722639681
    """

    name = "skew"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    stack_on = []
    stack_on_self = False
    description_template = "the skewness of {}"

    def get_function(self, agg_type=Library.PANDAS):
        return pd.Series.skew



from featuretools import config


def test_get_default_config_does_not_change():
    old_config = config.get_all()

    key = "primitive_data_folder"
    value = "This is an example string"
    config.set({key: value})
    config.set_to_default()

    assert config.get(key) != value

    config.set(old_config)


def test_set_and_get_config():
    key = "primitive_data_folder"
    old_value = config.get(key)
    value = "This is an example string"

    config.set({key: value})
    assert config.get(key) == value

    config.set({key: old_value})


def test_get_all():
    assert config.get_all() == config._data



from datetime import datetime

import pandas as pd

from featuretools.primitives import Season


class TestSeason:
    def test_regular(self):
        primitive_instance = Season()
        primitive_func = primitive_instance.get_function()
        case = pd.date_range(start="2019-01", periods=12, freq="m").to_series()
        answer = pd.Series(
            [
                "winter",
                "winter",
                "spring",
                "spring",
                "spring",
                "summer",
                "summer",
                "summer",
                "fall",
                "fall",
                "fall",
                "winter",
            ],
            dtype="string",
        )
        given_answer = primitive_func(case)
        pd.testing.assert_series_equal(
            given_answer.reset_index(drop=True),
            answer.reset_index(drop=True),
        )

    def test_nat(self):
        primitive_instance = Season()
        primitive_func = primitive_instance.get_function()
        case = pd.Series(
            [
                "NaT",
                "2019-02",
                "2019-03",
                "NaT",
            ],
        ).astype("datetime64[ns]")
        answer = pd.Series([pd.NA, "winter", "winter", pd.NA], dtype="string")
        given_answer = pd.Series(primitive_func(case))
        pd.testing.assert_series_equal(given_answer, answer)

    def test_datetime(self):
        primitive_instance = Season()
        primitive_func = primitive_instance.get_function()
        case = pd.Series(
            [
                datetime(2011, 3, 1),
                datetime(2011, 6, 1),
                datetime(2011, 9, 1),
                datetime(2011, 12, 1),
                # leap year
                datetime(2020, 2, 29),
            ],
        )
        answer = pd.Series(
            ["winter", "spring", "summer", "fall", "winter"],
            dtype="string",
        )
        given_answer = primitive_func(case)
        pd.testing.assert_series_equal(given_answer, answer)



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double

from featuretools.primitives.base import AggregationPrimitive


class Variance(AggregationPrimitive):
    """Calculates the variance of a list of numbers.

    Description:
        Given a list of numbers, return the variance,
        using numpy's built-in variance function. Nan
        values in a series will be ignored. Return nan
        when the series is empty or entirely null.

    Examples:
        >>> variance = Variance()
        >>> variance([0, 3, 4, 3])
        2.25

        Null values in a series will be ignored.

        >>> variance = Variance()
        >>> variance([0, 3, 4, 3, None])
        2.25
    """

    name = "variance"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = np.nan

    def get_function(self):
        return np.var



from featuretools import IdentityFeature
from featuretools.primitives.utils import PrimitivesDeserializer


def test_relationship_path(es):
    value = IdentityFeature(es["log"].ww["value"])
    assert len(value.relationship_path) == 0


def test_serialization(es):
    value = IdentityFeature(es["log"].ww["value"])

    dictionary = {
        "name": "value",
        "column_name": "value",
        "dataframe_name": "log",
    }

    assert dictionary == value.get_arguments()
    assert value == IdentityFeature.from_dictionary(
        dictionary,
        es,
        {},
        PrimitivesDeserializer,
    )



import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, EmailAddress

from featuretools.primitives.base import TransformPrimitive


class EmailAddressToDomain(TransformPrimitive):
    """Determines the domain of an email

    Description:
        EmailAddress input should be a string. Will return Nan
        if an invalid email address is provided, or if the input is
        not a string.

    Examples:
        >>> email_address_to_domain = EmailAddressToDomain()
        >>> email_address_to_domain(['name@gmail.com', 'name@featuretools.com']).tolist()
        ['gmail.com', 'featuretools.com']
    """

    name = "email_address_to_domain"
    input_types = [ColumnSchema(logical_type=EmailAddress)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        def email_address_to_domain(emails):
            # if the input is empty return an empty Series
            if len(emails) == 0:
                return pd.Series([], dtype="category")

            emails_df = pd.DataFrame({"email": emails})

            # if all emails are NaN expand won't propogate NaNs and will fail on indexing
            if emails_df["email"].isnull().all():
                emails_df["domain"] = np.nan
                emails_df["domain"] = emails_df["domain"].astype(object)
            else:
                # .str.strip() and .str.split() return NaN for NaN values and propogate NaNs into new columns
                emails_df["domain"] = (
                    emails_df["email"].str.strip().str.split("@", expand=True)[1]
                )
            return emails_df.domain.values

        return email_address_to_domain



import pandas as pd
import pytest
from woodwork.logical_types import (
    Categorical,
    Datetime,
    Double,
    Integer,
    NaturalLanguage,
)

from featuretools.entityset import EntitySet
from featuretools.tests.testing_utils import get_df_tags
from featuretools.utils.gen_utils import Library, import_or_none

dd = import_or_none("dask.dataframe")


@pytest.mark.skipif("not dd")
def test_add_dataframe(pd_es):
    dask_es = EntitySet(id="dask_es")
    log_dask = dd.from_pandas(pd_es["log"], npartitions=2)
    dask_es = dask_es.add_dataframe(
        dataframe_name="log_dask",
        dataframe=log_dask,
        index="id",
        time_index="datetime",
        logical_types=pd_es["log"].ww.logical_types,
        semantic_tags=get_df_tags(pd_es["log"]),
    )
    pd.testing.assert_frame_equal(
        pd_es["log"],
        dask_es["log_dask"].compute(),
        check_like=True,
    )


@pytest.mark.skipif("not dd")
def test_add_dataframe_with_non_numeric_index(pd_es, dask_es):
    df = pd.DataFrame({"id": ["A_1", "A_2", "C", "D"], "values": [1, 12, -34, 27]})
    dask_df = dd.from_pandas(df, npartitions=2)

    pd_es.add_dataframe(
        dataframe_name="new_dataframe",
        dataframe=df,
        index="id",
        logical_types={"id": Categorical, "values": Integer},
    )

    dask_es.add_dataframe(
        dataframe_name="new_dataframe",
        dataframe=dask_df,
        index="id",
        logical_types={"id": Categorical, "values": Integer},
    )

    pd.testing.assert_frame_equal(
        pd_es["new_dataframe"].reset_index(drop=True),
        dask_es["new_dataframe"].compute(),
    )


@pytest.mark.skipif("not dd")
def test_create_entityset_with_mixed_dataframe_types(pd_es, dask_es):
    df = pd.DataFrame({"id": [0, 1, 2, 3], "values": [1, 12, -34, 27]})
    dask_df = dd.from_pandas(df, npartitions=2)

    err_msg = (
        "All dataframes must be of the same type. "
        "Cannot add dataframe of type {} to an entityset with existing dataframes "
        "of type {}"
    )

    # Test error is raised when trying to add Dask dataframe to entityset with existing pandas dataframes
    with pytest.raises(
        ValueError,
        match=err_msg.format(type(dask_df), type(pd_es.dataframes[0])),
    ):
        pd_es.add_dataframe(
            dataframe_name="new_dataframe",
            dataframe=dask_df,
            index="id",
        )

    # Test error is raised when trying to add pandas dataframe to entityset with existing dask dataframes
    with pytest.raises(
        ValueError,
        match=err_msg.format(type(df), type(dask_es.dataframes[0])),
    ):
        dask_es.add_dataframe(dataframe_name="new_dataframe", dataframe=df, index="id")


@pytest.mark.skipif("not dd")
def test_add_last_time_indexes():
    pd_es = EntitySet(id="pd_es")
    dask_es = EntitySet(id="dask_es")

    sessions = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "user": [1, 2, 1, 3],
            "time": [
                pd.to_datetime("2019-01-10"),
                pd.to_datetime("2019-02-03"),
                pd.to_datetime("2019-01-01"),
                pd.to_datetime("2017-08-25"),
            ],
            "strings": ["I am a string", "23", "abcdef ghijk", ""],
        },
    )
    sessions_dask = dd.from_pandas(sessions, npartitions=2)
    sessions_logical_types = {
        "id": Integer,
        "user": Integer,
        "time": Datetime,
        "strings": NaturalLanguage,
    }

    transactions = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "session_id": [0, 0, 1, 2, 2, 3],
            "amount": [1.23, 5.24, 123.52, 67.93, 40.34, 50.13],
            "time": [
                pd.to_datetime("2019-01-10 03:53"),
                pd.to_datetime("2019-01-10 04:12"),
                pd.to_datetime("2019-02-03 10:34"),
                pd.to_datetime("2019-01-01 12:35"),
                pd.to_datetime("2019-01-01 12:49"),
                pd.to_datetime("2017-08-25 04:53"),
            ],
        },
    )
    transactions_dask = dd.from_pandas(transactions, npartitions=2)

    transactions_logical_types = {
        "id": Integer,
        "session_id": Integer,
        "time": Datetime,
        "amount": Double,
    }

    pd_es.add_dataframe(
        dataframe_name="sessions",
        dataframe=sessions,
        index="id",
        time_index="time",
    )
    dask_es.add_dataframe(
        dataframe_name="sessions",
        dataframe=sessions_dask,
        index="id",
        time_index="time",
        logical_types=sessions_logical_types,
    )

    pd_es.add_dataframe(
        dataframe_name="transactions",
        dataframe=transactions,
        index="id",
        time_index="time",
    )
    dask_es.add_dataframe(
        dataframe_name="transactions",
        dataframe=transactions_dask,
        index="id",
        time_index="time",
        logical_types=transactions_logical_types,
    )

    pd_es = pd_es.add_relationship("sessions", "id", "transactions", "session_id")
    dask_es = dask_es.add_relationship("sessions", "id", "transactions", "session_id")

    assert "foreign_key" in pd_es["transactions"].ww.semantic_tags["session_id"]
    assert "foreign_key" in dask_es["transactions"].ww.semantic_tags["session_id"]

    assert pd_es["sessions"].ww.metadata.get("last_time_index") is None
    assert dask_es["sessions"].ww.metadata.get("last_time_index") is None

    pd_es.add_last_time_indexes()
    dask_es.add_last_time_indexes()

    pd_lti_name = pd_es["sessions"].ww.metadata.get("last_time_index")
    spark_lti_name = dask_es["sessions"].ww.metadata.get("last_time_index")
    assert pd_lti_name == spark_lti_name
    pd.testing.assert_series_equal(
        pd_es["sessions"][pd_lti_name].sort_index(),
        dask_es["sessions"][spark_lti_name].compute().sort_index(),
        check_names=False,
    )


@pytest.mark.skipif("not dd")
def test_add_dataframe_with_make_index():
    values = [1, 12, -23, 27]
    df = pd.DataFrame({"values": values})
    dask_df = dd.from_pandas(df, npartitions=2)
    dask_es = EntitySet(id="dask_es")
    logical_types = {"values": Integer}
    dask_es.add_dataframe(
        dataframe_name="new_dataframe",
        dataframe=dask_df,
        make_index=True,
        index="new_index",
        logical_types=logical_types,
    )

    expected_df = pd.DataFrame({"values": values, "new_index": range(len(values))})
    pd.testing.assert_frame_equal(expected_df, dask_es["new_dataframe"].compute())


@pytest.mark.skipif("not dd")
def test_dataframe_type_dask(dask_es):
    assert dask_es.dataframe_type == Library.DASK



# -*- coding: utf-8 -*-
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable, NaturalLanguage

from featuretools.primitives.standard.transform.natural_language.count_string import (
    CountString,
)


class UpperCaseCount(CountString):
    """Calculates the number of upper case letters in text.

    Description:
        Given a list of strings, determine the number of characters in each string
        that are capitalized. Counts every letter individually, not just every
        word that contains capitalized letters.

        If a string is missing, return `NaN`

    Examples:
        >>> x = ['This IS a string.', 'This is a string', 'aaa']
        >>> upper_case_count = UpperCaseCount()
        >>> upper_case_count(x).tolist()
        [3.0, 1.0, 0.0]
    """

    name = "upper_case_count"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    default_value = 0

    def __init__(self):
        pattern = r"([A-Z])"
        super().__init__(string=pattern, is_regex=True, ignore_case=False)



import numpy as np
import pandas as pd

from featuretools.primitives import WhitespaceCount
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestWhitespaceCount(PrimitiveTestBase):
    primitive = WhitespaceCount

    def compare(self, primitive_initiated, test_cases, answers):
        primitive_func = primitive_initiated.get_function()
        primitive_answers = primitive_func(test_cases)
        return np.testing.assert_array_equal(answers, primitive_answers)

    def test_strings(self):
        x = pd.Series(
            ["", "hi im ethan!", "consecutive.    spaces.", " spaces-on-ends "],
        )
        answers = [0, 2, 4, 2]
        self.compare(self.primitive(), x, answers)

    def test_nan(self):
        x = pd.Series([np.nan, None, pd.NA, "", "This IS a STRING."])
        answers = [np.nan, np.nan, np.nan, 0, 3]
        self.compare(self.primitive(), x, answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)
